from phill.solver import Simul

params = Simul.create_default_params()
params.output.sub_directory = "examples_snek/tuto"
params.oper.nx = 6
params.oper.ny = 5
params.oper.nz = 4
params.oper.elem.order = params.oper.elem.order_out = 8

params.oper.nproc_min = 2

params.nek.general.num_steps = 10
params.nek.general.time_stepper = "bdf2"
params.nek.general.write_interval = 10

params.nek.stat.av_step = 2
params.nek.stat.io_step = 10

sim = Simul(params)

sim.make.exec("run_fg", nproc=2)



"""Base class for ``sim.output``.

"""
import inspect
import logging
import os
import pkgutil
import shutil
import stat
import textwrap
import warnings
from importlib import resources
from itertools import chain
from pathlib import Path
from socket import gethostname

import yaml
from inflection import underscore

from fluiddyn.io import stdout_redirected
from fluiddyn.util import mpi
from fluidsim_core.output import OutputCore
from fluidsim_core.params import iter_complete_params
from snek5000 import __version__, get_snek_resource, logger
from snek5000.make import _Nek5000Make
from snek5000.params import _save_par_file
from snek5000.solvers import get_solver_package, is_package
from snek5000.util import docstring_params
from snek5000.util.files import bisect_nek_files_by_time
from snek5000.util.smake import append_debug_flags, set_compiler_verbosity

from . import _make_path_session


class MissingConfigFilter:
    def filter(self, record):
        msg = record.msg
        if hasattr(msg, "startswith") and msg.startswith(
            "Missing a configuration file"
        ):
            if hasattr(self, "emitted") and self.emitted:
                return False
            self.emitted = True
        return True


missing_config_filter = MissingConfigFilter()
logger.addFilter(missing_config_filter)


class Output(OutputCore):
    """Container and methods for getting paths of and copying case files.

    Some important methods:

    - :meth:`snek5000.output.base.Output.get_path_solver_package` points to the
      directory containing the case files.
    - :meth:`snek5000.output.base.Output.get_paths` populates a list of
      files to copy.

    and important attributes:

    .. autoattribute:: name_solver
       :noindex:

        Initialized from ``sim.info.solver.short_name`` used to discover source
        code files, such as usr, box, par files. The value of ``name_solver``
        is also used to identify the entrypoint pointing to the solver module.
        Have a look at the :ref:`packaging tutorial <packaging>`.

    .. autoattribute:: package

        Initialized using :func:`snek5000.solvers.get_solver_package` and
        :attr:`name_solver`, points to the path to the package to discover
        source code files.

    .. autoattribute:: path_run
       :noindex:

        Path to the generated simulation directory.

    .. autoattribute:: path_session

        Path to subdirectory under :attr:`path_run` which would contain the
        field files upon execution. This path would be written into the
        `SESSION.NAME` file.

    """

    _config_filename = "config_simul.yml"

    @property
    def excludes(self):
        """Prefixes and suffixes of files which should be excluded from being
        copied."""
        return {
            "prefix": "__",
            "suffix": (".vimrc", ".tar.gz", ".o", ".py", ".usr.f", ".par.cfg"),
        }

    @property
    def makefile_usr_sources(self):
        """
        Sources for inclusion to makefile_usr.inc
        Dict[directory]  -> list of source files
        """
        return {
            # "source_directory": [
            #    (src1a.f, src1b.f)
            #    (src2a.f, src2b.f, src2c.f), ...
            # ]
        }

    @property
    def makefile_usr_obj(self):
        """Object files to be included in compilation. Should be exported as USR
        environment variable.

        """
        makefile_usr_obj = [
            sources[0].replace(".f", ".o")
            for sources in chain.from_iterable(self.makefile_usr_sources.values())
        ]
        return makefile_usr_obj

    @property
    def fortran_inc_flags(self):
        return (f"-I{inc_dir}" for inc_dir in self.makefile_usr_sources)

    @classmethod
    def _set_info_solver_classes(cls, classes):
        """Set the the classes for info_solver.classes.Output"""

        classes._set_child(
            "PrintStdout",
            dict(
                module_name="snek5000.output.print_stdout",
                class_name="PrintStdOut",
            ),
        )
        classes._set_child(
            "PhysFields",
            dict(module_name="snek5000.output.phys_fields", class_name="PhysFields"),
        )

        classes._set_child(
            "HistoryPoints",
            dict(
                module_name="snek5000.output.history_points", class_name="HistoryPoints"
            ),
        )

    @classmethod
    def _complete_info_solver(cls, info_solver):
        """Complete the info_solver instance with child class details (module
        and class names).

        """
        classes = info_solver.classes.Output._set_child("classes")
        cls._set_info_solver_classes(classes)
        # iteratively call _complete_info_solver of the above classes
        info_solver.classes.Output.complete_with_classes()

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""

        # Bare minimum
        attribs = {
            "HAS_TO_SAVE": True,
            "sub_directory": "",
            "session_id": 0,
        }
        params._set_child("output", attribs=attribs)
        params.output._set_doc(
            textwrap.dedent(
                """
    - ``HAS_TO_SAVE``: bool (default: True) If False, nothing new is saved in
      the directory of the simulation.
    - ``sub_directory``: str (default: "") A name of a sub-directory (relative
      to $FLUIDDYN_PATH_SCRATCH) wherein the directory of the simulation
      (``path_run``) is saved.
    - ``session_id``: int (default: 0) Determines the sub-directory,
      ``path_session`` in which the field files would be generated during
      runtime. The session directory takes the form `session_{session_id}`.

    .. note::

        In short, the field files would be generated under
        ``$FLUIDDYN_PATH_SCRATCH/<path_run>/<path_session>``

"""
            )
        )

        dict_classes = info_solver.classes.Output.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    @classmethod
    def get_path_solver_package(cls):
        """Get the path towards the solver package."""
        return Path(inspect.getmodule(cls).__file__).parent

    @classmethod
    def find_configfile(cls, host=None):
        """Get path of the Snakemake configuration file for the current machine.
        All configuration files are stored under ``etc`` sub-package.

        Parameters
        ----------
        host: str
            Override hostname detection and specify it instead

        """
        if not host:
            host = os.getenv(
                "SNIC_RESOURCE", os.getenv("GITHUB_WORKFLOW", gethostname())
            )
        path_solver_package = cls.get_path_solver_package()
        xdg_config = Path(
            os.path.expandvars(os.getenv("XDG_CONFIG_HOME", "$HOME/.config"))
        )
        configfile_root = path_solver_package / "etc" / f"{host}.yml"
        configfile_xdg_config_host = xdg_config / f"snek5000/{host}.yml"
        configfile_xdg_config = xdg_config / "snek5000.yml"
        configfile_default = Path(get_snek_resource("default_configfile.yml"))

        custom_configfiles = (
            configfile_xdg_config_host,
            configfile_xdg_config,
            configfile_root,
        )

        for configfile in custom_configfiles:
            if configfile.exists():
                break
        else:
            configfile = configfile_default
            logger.warning(
                (
                    "Missing a configuration file describing compilers and "
                    "flags. Create one at either of the following paths to "
                    "avoid future warnings:\n"
                )
                + "\n".join(map(str, custom_configfiles))
                + "\nThe command `snek-generate-config` could be used to create "
                "a user config file for you."
                f"\nUsing default configuration for now:\n{configfile}"
            )

        return configfile

    @classmethod
    def update_snakemake_config(
        cls, config, name_solver, /, verbosity=0, env_sensitive=None, **kwargs
    ):
        """Update snakemake config in-place with name of the solver / case,
        path to configfile and compiler flags

        Parameters
        ----------
        config: dict
            Snakemake configuration
        name_solver: str
            Short name of the solver, also known as case name
        verbosity: int
            Set compiler verbosity level. See :func:`snek5000.util.smake.set_compiler_verbosity`
        env_sensitive: bool (None)
            If ``False``, the ``config`` dictionary is not modified (allows for
            reproducible runs). If ``True``, the ``config`` dictionary is
            modified based on environment variables. If ``None`` (default), the
            value of ``env_sensitive`` is obtained with
            ``os.environ.get("SNEK_UPDATE_CONFIG_ENV_SENSITIVE", False)``.

        .. deprecated:: 0.8.0

            The ``warnings`` parameter is deprecated! Use ``verbosity=0`` (now default)
            to disable warnings. If you need ``warnings=True``, similar behaviour can
            be obtained by ``verbosity=1`` or ``verbosity=2``.

        """
        mandatory_config = {
            "CC",
            "FC",
            "MPICC",
            "MPIFC",
            "MPIEXEC",
            "MPIEXEC_FLAGS",
            "CFLAGS",
            "FFLAGS",
        }
        missing_config = mandatory_config - set(config)

        if missing_config:
            raise ValueError(
                f"Some keys are missing from the configfile "
                f"{cls.get_configfile()}: {missing_config}"
            )

        try:
            # Suppress warnings for not instantiating Output with sim or params
            logging_level = logger.getEffectiveLevel()
            logger.setLevel(logging.ERROR)
            temp = cls()
        finally:
            logger.setLevel(logging_level)

            config.update(
                {
                    "CASE": name_solver,
                    "file": Path(cls._config_filename).resolve(),
                    "includes": " ".join(temp.fortran_inc_flags),
                    "objects": " ".join(temp.makefile_usr_obj),
                }
            )

            if "warnings" in kwargs:
                warnings.warn(
                    "Parameter warnings is deprecated, use ``verbosity`` instead/",
                    DeprecationWarning,
                )
                verbosity = int(kwargs["warnings"])

            set_compiler_verbosity(config, verbosity)
            append_debug_flags(config)

            if env_sensitive is None:
                env_sensitive = os.environ.get(
                    "SNEK_UPDATE_CONFIG_ENV_SENSITIVE", False
                )
                if isinstance(env_sensitive, str):
                    # correct for "0", "false", "False"
                    env_sensitive = bool(yaml.safe_load(env_sensitive))

            if env_sensitive:
                logger.info(
                    "env_sensitive = True => attempting to update config from environment variables."
                )
                config.update(
                    {
                        key: os.getenv(key, original_value)
                        for key, original_value in config.items()
                    }
                )

    def __init__(self, sim=None, params=None):
        self.sim = sim
        try:
            self.name_solver = sim.info.solver.short_name
        except AttributeError:
            pass
        else:
            self.package = get_solver_package(self.name_solver)

        self.path_solver_package = self.get_path_solver_package()

        if sim:
            self.oper = sim.oper
            self.params = sim.params.output
            # Same as package name __name__
            super().__init__(sim)
        elif params:
            # At least initialize params
            self.params = params.output
        else:
            self.params = None
            logger.warning(
                "Initializing Output class without sim or params might lead to errors."
            )

        self.path_session = self._init_path_session()

        if sim:
            # initialize objects
            dict_classes = sim.info.solver.classes.Output.import_classes()
            for cls_name, Class in dict_classes.items():
                # only initialize if Class is not the Output class
                if not isinstance(self, Class):
                    obj_name = underscore(cls_name)
                    setattr(self, obj_name, Class(self))
                    self.sim._objects_to_print += "{:28s}{}\n".format(
                        f"sim.output.{obj_name}: ", Class
                    )

    def _init_path_session(self):
        """Initialize :attr:`path_session` and ``params.output.path_session``
        from ``params.output.session_id``. Unlike :meth:`_init_path_run`, the
        directory will not be created.

        Returns
        -------
        path_session: path-like

        """
        if not self.params or not hasattr(self, "path_run"):
            logger.debug("Attribute sim.output.path_session will not be initialized.")
            return None

        try:
            session_id = self.params.session_id
        except AttributeError:
            # For compatibility while loading old simulations
            path_session = Path(self.path_run)
            logger.warning(
                "Parameter params.output.session_id is undefined. "
                "Attribute sim.output.path_session is set as sim.output.path_run."
            )
        else:
            path_session = _make_path_session(self.path_run, session_id)

        self.params._set_attrib("path_session", path_session)

        return path_session

    def _init_sim_repr_maker(self):
        """Adds mesh description to name of the simulation. Called by the
        ``_init_name_run`` method"""
        sim_repr_maker = super()._init_sim_repr_maker()
        self.oper._modify_sim_repr_maker(sim_repr_maker)

        return sim_repr_maker

    def _get_resources(self, package=None):
        """Get a generator of resources (files) in a package, excluding
        directories (subpackages).

        :returns: generator

        """
        excludes = self.excludes
        if not package:
            package = self.package

        try:
            contents_pkg = resources.contents(package)
        except ImportError:
            raise FileNotFoundError(
                f"Cannot resolve subpackage name_solver={package} "
                f"at path_solver_package={self.path_solver_package}"
            )

        return (
            f
            for f in contents_pkg
            if (
                resources.is_resource(package, f)
                and not any(f.startswith(ext) for ext in excludes["prefix"])
                and not any(f.endswith(ext) for ext in excludes["suffix"])
            )
        )

    def _get_subpackages(self):
        """Get a dictionary of subpackages with values generated by
        :meth:`_get_resources`.

        :returns: dict

        """
        root = self.path_solver_package
        subpackages = {
            subpkg.name.replace(f"{root.name}.", ""): self._get_resources(subpkg.name)
            for subpkg in pkgutil.walk_packages([str(root)], prefix=f"{self.package}.")
            if is_package(subpkg)
        }

        return subpackages

    def get_paths(self):
        """Get a list of paths to all case files.

        :returns: list

        """
        paths = []

        # abl.usr -> /path/to/abl/abl.usr
        paths += [
            self.path_solver_package / resource for resource in self._get_resources()
        ]

        for subpkg, res in self._get_subpackages().items():
            # toolbox -> /path/to/abl/toolbox
            subpkg_root = self.path_solver_package / subpkg.replace(".", os.sep)
            # main.f -> /path/to/abl/toolbox/main.f
            paths += [subpkg_root / resource for resource in res]
        return paths

    def copy(self, new_dir, force=False):
        """Copy case files to a new directory. The directory does not have to be present.

        :param new_dir: A str or Path-like instance pointing to the new directory.
        :param force: Force copy would overwrite if files already exist.

        """
        # Avoid race conditions! Should be only executed by rank 0.
        if mpi.rank != 0:
            return

        abs_paths = self.get_paths()
        subpackages = self._get_subpackages()

        path_solver_package = self.path_solver_package

        def conditional_ignore(src, names):
            """Ignore if not found in ``abs_paths``."""
            src = Path(src)
            include = abs_paths + [
                path_solver_package / subpkg for subpkg in subpackages
            ]
            exclude = tuple(
                name
                for name in names
                if not any((src / name) == path for path in include)
            )

            logger.debug(
                "".join(
                    (
                        f"- src: {src}",
                        "\n- include:",
                        " ".join(Path(i).name for i in include),
                        "\n- exclude:",
                        " ".join(Path(i).name for i in exclude),
                    )
                )
            )
            return exclude

        new_root = Path(new_dir)
        # `dirs_exist_ok`` new in Python 3.8
        shutil.copytree(
            src=path_solver_package,
            dst=new_root,
            symlinks=False,
            ignore=conditional_ignore,
            dirs_exist_ok=True,
        )

        # special case for .usr.f: copy to .usr
        paths_usr_f = list(path_solver_package.glob("*.usr.f"))
        for path_usr_f in paths_usr_f:
            shutil.copyfile(path_usr_f, new_root / path_usr_f.stem)

    def write_box(self, template):
        """Write <case name>.box file from box.j2 template.

        .. seealso::
            :ref:`nek:mesh_gen`

        """
        if mpi.rank == 0:
            box_file = self.sim.path_run / f"{self.name_solver}.box"
            logger.info(f"Writing box file... {box_file}")
            with open(box_file, "w") as fp:
                self.sim.oper.write_box(
                    template, fp, comments=self.sim.params.short_name_type_run
                )

    def write_size(self, template):
        """Write SIZE file from SIZE.j2 template.

        .. seealso::

            Nek5000 docs on :ref:`nek:case_files_size`

        """
        if mpi.rank == 0:
            size_file = self.sim.path_run / "SIZE"
            logger.info(f"Writing SIZE file... {size_file}")
            with open(size_file, "w") as fp:
                self.oper.write_size(
                    template, fp, comments=self.sim.params.short_name_type_run
                )

    def write_makefile_usr(self, template, fp=None, **template_vars):
        """Write the makefile_usr.inc file which gets included in the main
        makefile by the ``makenek`` tool.

        Parameters
        ----------
        template : jinja2.environment.Template
            Template instance loaded from something like ``makefile_usr.inc.j2``
        fp : io.TextIOWrapper
            File handler to write to
        comments: str
            Comments on top of the box file
        template_vars: dict
            Keyword arguments passed while rendering the Jinja templates

        """

        paths_of_sources = []

        for path_dir, list_of_sources in self.makefile_usr_sources.items():
            for sources in list_of_sources:
                paths_of_sources.append([f"{path_dir}/{file}" for file in sources])

        if mpi.rank == 0:
            comments = "Autogenerated using snek5000.output.Output.write_makefile_usr\n"
            if self.sim is not None:
                comments += self.sim.params.short_name_type_run

            template_vars.update(
                {"list_of_sources": paths_of_sources, "comments": comments}
            )
            output = template.render(**template_vars)
            if fp:
                fp.write(output)
            else:
                makefile_usr = self.sim.path_run / "makefile_usr.inc"
                with open(makefile_usr, "w") as fp:
                    fp.write(output)

    def write_snakemake_config(self, custom_env_vars=None, host=None):
        """Write the config file in the simulation directory

        Parameters
        ----------
        custom_env_vars: dict (None)
            Environment variables used to update the configuration found by
            :meth:`find_configfile`.
        host: str
            Override hostname detection and specify it instead

        """
        if mpi.rank != 0:
            return

        path_configfile = self.find_configfile(host=host)
        path_configfile_simul = self.sim.path_run / self._config_filename

        with open(path_configfile) as file:
            config = yaml.safe_load(file)

        if custom_env_vars is None:
            shutil.copyfile(path_configfile, path_configfile_simul)
        else:
            config.update(custom_env_vars)
            with open(path_configfile_simul, "w") as file:
                yaml.dump(config, file)

        return config

    @staticmethod
    def build_nek5000(config):
        """Build Nek5000, if needed. This method is automatically invoked
        during :meth:`post_init`.

        Examples
        --------
        If compiler configuration is changed via a script after Simulation
        initialization, a rebuild can be manually triggered as follows:

        >>> config = sim.output.write_snakemake_config(
        ...     custom_env_vars={"CFLAGS": "-O0 -g", FFLAGS": "-O0 -g"}
        ... )
        >>> sim.output.build_nek5000(config)

        """
        nek5000 = _Nek5000Make()
        if not nek5000.build(config):
            raise RuntimeError("Nek5000 build failed.")

    @staticmethod
    def write_compile_sh(template, config, fp=None, path=None):
        """Write a standalone ``compile.sh`` shell script to compile the user
        code.

        Parameters
        ----------
        template: jinja2.environment.Template
            Template similar to ``snek5000/resources/compiler_sh.j2``
        config: dict
            Snakemake configuration
        fp: io.TextIOWrapper
            File pointer
        path: str or Path
            Path to write the file to

        """
        output = template.render(
            INC=config["includes"],
            USR=config["objects"],
            **config,
        )
        path = str(path)
        if fp:
            fp.write(output)
        elif path:
            with open(path, "w") as fp:
                fp.write(output)
        else:
            raise ValueError("Either file pointer or the path to it must be provided.")

        if path:
            os.chmod(path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH)

    def get_field_file(self, prefix="", index=-1, t_approx=None):
        """Get a field file from ``path_session``.

        Parameters
        ----------
        prefix: str
            Prefix for special field files; for examples KTH statistics files use prefix `sts`.

        index: int
            Index to match a specific field file. If index > 0, the file
            extension is matched as ``{prefix}case0.f{index:05d}``. If index <
            0, the file is indexed from the end of a list of files

        t_approx: float
            Find a file from approximate simulation time

        Returns
        -------
        file: Path

        """
        case = self.name_solver
        path_session = self.sim.output.path_session

        if index > 0 and t_approx:
            raise ValueError("Specify either index or t_approx at a time, not both.")
        elif index > 0:
            pattern = f"{prefix}{case}0.f{index:05d}"
            file = path_session / pattern
            if file.exists():
                return file
            else:
                logger.warning(
                    f"{file} not found. Attempting to index a file from a "
                    "sorted list of field files"
                )
        elif t_approx:
            index = slice(None)

        pattern = f"{prefix}{case}0.f?????"
        try:
            result = sorted(path_session.glob(pattern))[index]
            if t_approx:
                result = bisect_nek_files_by_time(result, t_approx)

        except IndexError as err:
            raise FileNotFoundError(
                f"Cannot {index =} / find {t_approx =} in {path_session}/{pattern} "
            ) from err
        else:
            return result

    def post_init(self):
        """Logs info on instantiated classes and finally :meth:`copy` all
        source code to simulation directory

        """

        if mpi.rank == 0:
            print(f"path_run: {self.path_run}")
            logger.info(f"session_id: {self.params.session_id}")

        # This also calls _save_info_solver_params_xml
        with stdout_redirected():
            # We gather objects to print within Snek5000
            super().post_init()

        logger.info(self.sim._objects_to_print)

        # Write source files to compile the simulation
        if mpi.rank == 0 and self._has_to_save and self.sim.params.NEW_DIR_RESULTS:
            self.copy(self.path_run)
            config = self.write_snakemake_config()
            self.build_nek5000(config)
            self.post_init_create_additional_source_files()

    def post_init_create_additional_source_files(self):
        """Create the .box, SIZE and makefile_usr files from their template"""
        for name in ("box", "size", "makefile_usr"):
            try:
                template = getattr(self, f"template_{name}")
            except AttributeError:
                pass
            else:
                if template is not None:
                    getattr(self, f"write_{name}")(template)

    def _save_info_solver_params_xml(self, replace=False):
        """Saves the par file, along with ``params_simul.xml`` and
        ``info_solver.xml``"""
        params = self.sim.params
        if mpi.rank == 0:
            par_file = Path(self.path_run) / f"{self.name_solver}.par"

            if self._has_to_save and params.NEW_DIR_RESULTS:
                logger.info(
                    f"Writing params files... {par_file}, params_simul.xml, "
                    "info_solver.xml"
                )
                _save_par_file(params, par_file, mode="x")
            elif self._has_to_save:
                logger.info(f"Updating {par_file}, params_simul.xml")
                _save_par_file(params, par_file)

                # Update params_simul.xml here, since FluidSim Core will only
                # do it if NEW_DIR_RESULTS = True
                params_xml_path = self.path_run / "params_simul.xml"
                params_xml_path.unlink()
                comment = f"""\
This file should not be modified (except for adding xml comments).
Created by the Python programs:
snek5000 {__version__}
"""

                params._save_as_xml(path_file=params_xml_path, comment=comment)

        super()._save_info_solver_params_xml(replace, comment=f"snek5000 {__version__}")


Output.__doc__ += """
    Notes
    -----
    Here, only the documention for ``params.output`` is displayed.

    .. seealso::

        - ``params.oper`` at :mod:`snek5000.operators`
        - ``params.nek`` at :mod:`snek5000.solvers.base`

""" + docstring_params(
    Output
)



from io import StringIO


def test_init(oper):
    print(oper.produce_str_describing_oper())
    print(oper.produce_long_str_describing_oper())


def test_properties(oper):
    assert oper.max_n_seq == 1024
    assert oper.max_n_loc == 171


def test_box_template(oper, jinja_env, datadir):
    box = jinja_env.get_template("box.j2")
    expected = (datadir / "test_box_template.box").read_text()

    with StringIO() as buffer:
        oper.write_box(box, fp=buffer, comments=__name__)
        assert buffer.getvalue() == expected


def test_box_2d(oper2d, jinja_env, datadir):
    box = jinja_env.get_template("box.j2")
    expected = (datadir / "test_box_2d.box").read_text()

    with StringIO() as buffer:
        oper2d.write_box(box, fp=buffer, comments=__name__)
        assert buffer.getvalue() == expected


def test_size_template(sim, jinja_env, datadir):
    size = jinja_env.get_template("SIZE.j2")
    expected = (datadir / "test_size_template.f").read_text()

    with StringIO() as buffer:
        sim.oper.write_size(size, fp=buffer, comments=__name__)
        assert buffer.getvalue().splitlines() == expected.splitlines()


def test_size_2d(sim2d, jinja_env, datadir):
    size = jinja_env.get_template("SIZE.j2")
    expected = (datadir / "test_size_2d.f").read_text()

    with StringIO() as buffer:
        sim2d.oper.write_size(size, fp=buffer, comments=__name__)
        assert buffer.getvalue().splitlines() == expected.splitlines()


def test_sim_path_run(sim):
    params = sim.params
    assert params.oper.Lx == params.oper.Ly == params.oper.Lz == 1.0
    assert sim.oper.produce_str_describing_oper() in sim.name_run



import bisect
import os
import re
from pathlib import Path
from shutil import copy2

from pymech.neksuite.field import read_header

from fluiddyn.io import FLUIDSIM_PATH

from .. import logger
from ..params import load_params


def _is_empty_directory(path):
    path = Path(path)
    if not path.is_dir():
        return False
    return not bool(os.listdir(path))


def next_path(old_path, force_suffix=False, return_suffix=False):
    """Generate a new path with an integer suffix

    Parameters
    ----------
    old_path: str or path-like
        Path to check for existence

    force_suffix:
        If true, will not check if the `old_path` can be used and adds
        a suffix in the end.

    return_suffix:
        If true, returns the integer suffix along with the path/

    Returns
    -------
    (i, new_path): tuple[int, Path]
        If `return_suffix` is `True`.

    new_path: Path
        A path (with an integer suffix) which does not yet exist in the
        filesystem.

    Examples
    --------
    >>> import os
    >>> os.chdir("/tmp")

    >>> next_path("test.txt")  # path does not exist
    PosixPath('test.txt')

    >>> next_path("test.txt", force_suffix=True)  # path does not exists
    PosixPath('test_00.txt')

    >>> Path("test.txt").touch()
    >>> next_path("test.txt")  # path exists
    PosixPath('test_00.txt')

    >>> Path("test_00.txt").touch()
    >>> next_path("test.txt")  # path and the next one both exists
    PosixPath('test_01.txt')

    >>> Path("test.txt").unlink()  # cleanup
    >>> Path("test_00.txt").unlink()

    """

    def int_suffix(p, integer):
        stem = p.stem
        # for example: remove .tar from the end, if any
        for suffix in p.suffixes:
            stem = re.sub(f"{suffix}$", "", stem)

        return p.parent / "".join([stem, f"_{integer:02d}", *p.suffixes])

    old_path = Path(old_path)

    if not force_suffix and not (
        old_path.exists() and not _is_empty_directory(old_path)
    ):
        return old_path

    i = 0
    new_path = int_suffix(old_path, i)
    while new_path.exists() and not _is_empty_directory(new_path):
        logger.debug(f"Path exists: {new_path}")
        i += 1
        new_path = int_suffix(old_path, i)

    logger.debug(f"Next path available: {new_path}")

    return (i, new_path) if return_suffix else new_path


def create_session(case, re2, ma2, par):
    """Creates a session and write the path to a `SESSION.NAME` file.
    Then, symlinks re2 and ma2 files, and copies the par file.

    Parameters
    ----------
    case : str
        Case name
    re2 : str
        Mesh file name
    ma2 : str
        Connectivity mapping file name
    par : str
        Parameter file name

    """
    params = load_params()
    session_dir = Path(params.output.path_session).relative_to(params.path_run)

    session_dir.mkdir(exist_ok=True)

    with open("SESSION.NAME", "w") as session_name:
        # use relative paths to avoid 132 character limit in Nek5000
        session_name.write(f"{case}\n./{session_dir}\n")

    for file in (re2, ma2):
        # use relative symlinks
        (session_dir / file).symlink_to(f"../{file}")

    # Copy par files to run without recompiling
    copy2(par, session_dir / par)


class LazyNekFile:
    """A small data stucture to assist bisection sort by simulation time,
    :func:`bisect_nek_files_by_time`

    """

    def __init__(self, path):
        self.path = path

    @property
    def time(self):
        return read_header(self.path).time

    def __gt__(self, other):
        time = other.time if isinstance(other, type(self)) else other
        return self.time > time

    def __lt__(self, other):
        time = other.time if isinstance(other, type(self)) else other
        return self.time < time

    def __repr__(self):
        return f"LazyNekFile <{self.path}>"


def bisect_nek_files_by_time(files, time):
    """Bisect Nek5000 field files by the simulation time metadata. The
    bisection search is done using :func:`bisect.bisect_left`

    Parameters
    ----------
    files: iterable of str or path-like
        List of file names / paths.

    time: float
        Approximate simulation time. Finds ``file`` such that
        :func:`read_header(file).time <pymech:pymech.neksuite.field.read_header>` < ``time``.

    Returns
    -------
    file: str or path-like
        Path to field file.

    """
    files = sorted(files)
    lazy_files = [LazyNekFile(file) for file in files]

    index = min(bisect.bisect_left(lazy_files, time), len(files) - 1)

    return files[index]


def _path_try_from_fluidsim_path(path_dir):
    """Converts to a :class:`pathlib.Path` object and if it does not exists,
    attempts a path relative to environment variable ``FLUIDSIM_PATH``.

    """
    path = Path(path_dir)

    if not path.exists():
        logger.info("Trying to open the path relative to $FLUIDSIM_PATH")
        path = Path(FLUIDSIM_PATH) / path_dir

    return path



from snek5000_tgv.solver import Simul

params = Simul.create_default_params()
params.output.sub_directory = "examples_snek/tuto"

params.oper.nx = params.oper.ny = params.oper.nz = 8
params.oper.elem.order = params.oper.elem.order_out = 8
params.oper.nproc_min = 2

params.nek.velocity.residual_tol = 1e-07
params.nek.pressure.residual_tol = 1e-05

params.nek.general.end_time = 10
params.nek.general.dt = -1
params.nek.general.target_cfl = 1.4
params.nek.general.extrapolation = "OIFS"

params.nek.general.write_control = "runTime"
params.nek.general.write_interval = 2.0

params.output.spatial_means.write_interval = 0.5

sim = Simul(params)
sim.make.exec("run_fg", nproc=2)



"""Calculate physical constants"""
import math


def coriolis_freq(lat):
    r"""Compute coriolis frequency in the f-plane.

    .. math::

        \omega &= 7.2921150 ± 0.0000001×10^{−5} \, \text{radians per second} \\
        f &= 2 \omega \sin(\theta)


    .. seealso::
        https://en.wikipedia.org/wiki/Earth%27s_rotation#Angular_speed

    Parameters
    ----------
    lat: float
        Latitude in degrees, which gets converted to radians :math:`\theta`.

    Returns
    -------
    f: float
        Coriolis frequency in radians per second

    """
    omega = 7.2921150e-5
    theta = math.radians(lat)
    f = 2 * omega * math.sin(theta)
    return f



import builtins
import os
import signal
from time import perf_counter, sleep

import numpy as np
from scipy.signal import argrelmax
from snek5000_cbox.solver import Simul

from fluiddyn.util import time_as_str

params = Simul.create_default_params()

aspect_ratio = 1.0
params.prandtl = 0.71

# The onset of oscillatory flow for aspect ratio 1.0 is at Ra_c = 1.825e8
params.Ra_side = 2e8

params.output.sub_directory = "examples_snek/tuto"

params.oper.dim = 2
params.oper.nproc_min = 2

nb_elements = ny = 8
params.oper.ny = nb_elements
nx = params.oper.nx = int(nb_elements / aspect_ratio)

Ly = params.oper.Ly
Lx = params.oper.Lx = Ly / aspect_ratio

order = params.oper.elem.order = params.oper.elem.order_out = 8

params.oper.mesh_stretch_factor = 0.1  # zero means regular

params.short_name_type_run = f"Ra{params.Ra_side:.3e}_{nx*order}x{ny*order}"

# creation of the coordinates of the points saved by history points
n1d = 5
small = Lx / 10

xs = np.linspace(0, Lx, n1d)
xs[0] = small
xs[-1] = Lx - small

ys = np.linspace(0, Ly, n1d)
ys[0] = small
ys[-1] = Ly - small

coords = [(x, y) for x in xs for y in ys]

params.output.history_points.coords = coords
params.oper.max.hist = len(coords) + 1

params.nek.velocity.residual_tol = 1e-08
params.nek.pressure.residual_tol = 1e-05

params.nek.general.end_time = 1000
params.nek.general.stop_at = "endTime"
params.nek.general.target_cfl = 2.0
params.nek.general.time_stepper = "BDF3"
params.nek.general.extrapolation = "OIFS"

params.nek.general.write_control = "runTime"
params.nek.general.write_interval = 50

params.output.history_points.write_interval = 10


if __name__ == "__main__":
    sim = Simul(params)

    # to save the PID of the simulation (a number associated with the process)
    sim.output.write_snakemake_config(
        custom_env_vars={"MPIEXEC_FLAGS": "--report-pid PID.txt --oversubscribe"}
    )
    # run the simulation in the background (non blocking call)
    sim.make.exec("run", nproc=2)

    pid_file = sim.path_run / "PID.txt"

    # we wait for the simulation to be started and the PID file to be created
    n = 0
    while not pid_file.exists():
        sleep(1)
        n += 1
        if n > 60:
            raise RuntimeError(f"{pid_file} does not exist.")

    with open(pid_file) as file:
        pid = int(file.read().strip())

    # we know the PID of the simulation so we can control it!

    # a simple way to print in stdout and in a file
    path_log_py = sim.path_run / f"log_py_{time_as_str()}.txt"

    def print(*args, sep=" ", end="\n", **kwargs):
        builtins.print(*args, **kwargs)
        with open(path_log_py, "a") as file:
            file.write(sep.join(str(arg) for arg in args) + end)

    print(f"{pid = }")

    def check_running():
        """Check for the existence of a running process"""
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    while check_running():
        sleep(2)
        t0 = perf_counter()
        coords, df = sim.output.history_points.load_1point(
            index_point=12, key="temperature"
        )
        t_last = df.time.max()
        print(
            f"{time_as_str()}, {t_last = :.2f}: "
            f"history_points loaded in {perf_counter() - t0:.2f} s"
        )

        if t_last < 500:
            continue

        temperature = df.temperature.to_numpy()
        times = df.time.to_numpy()

        temperature = temperature[times > 500]
        times = times[times > 500]

        # we know that there are oscillations growing exponentially
        # we look for the positive maxima of the signal
        indices_maxima = argrelmax(temperature)
        temp_maxima = temperature[indices_maxima]
        temp_maxima = temp_maxima[temp_maxima > 0]

        # similar to a second order derivative
        diff_maxima = np.diff(np.diff(temp_maxima))

        if any(diff_maxima < 0):
            # as soon as the growth starts to saturate
            print(
                f"Saturation of the instability detected at t = {t_last}\n"
                "Terminate simulation."
            )

            os.kill(pid, signal.SIGTERM)
            break



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import os
import runpy
import subprocess
import sys
from datetime import date
from pathlib import Path
from subprocess import PIPE

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata

import breathe

import snek5000
from snek5000 import util
from snek5000.config import ensure_config_file

ensure_config_file()


def root(module):
    return os.fspath(Path(module.__file__).parent)


sys.path.insert(0, root(breathe))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, root(snek5000))

NEK_SOURCE_ROOT = snek5000.get_nek_source_root()
print("NEK_SOURCE_ROOT = ", NEK_SOURCE_ROOT)
print("sys.path =\n   ", "\n    ".join(sys.path))

# -- Project information -----------------------------------------------------

project = "snek5000"
_meta = metadata.metadata(project)
_today = date.today()
copyright = (
    f"2019 - {_today.year}, Ashwin Vishnu Mohanan. Published: {_today.isoformat()}"
)
author = "Ashwin Vishnu Mohanan"

version = ".".join(snek5000.__version__.split(".")[:3])
# The full version, including alpha/beta/rc tags
release = snek5000.__version__

_py_min_version = _meta.get("Requires-Python").split(">=")[-1]

rst_prolog = f"""
.. |author| replace:: {author}
.. |today| replace:: {_today}
.. |py_min_version| replace:: {_py_min_version}
"""

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_inline_tabs",
    "IPython.sphinxext.ipython_console_highlighting",
    # "recommonmark",
    # "myst_parser",
    "myst_nb",
    "sphinx_copybutton",
]

# Execute ipynb files into with a cache ...
nb_execution_mode = "cache"
# ... except these ipynb files
nb_execution_excludepatterns = ["**/*.ipynb", "debug/**/*"]
nb_execution_raise_on_error = True
nb_execution_show_tb = True
nb_execution_timeout = 600
nb_merge_streams = True

# CSS selector which modifies the sphinx-copybutton feature
copybutton_selector = ",".join(
    [
        f"div.highlight-{css_class} div.highlight pre"
        for css_class in ("python", "ipython3", "sh", "ini", "default")
    ]
)

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst.md": "myst-nb",
}

# Execute Doxygen
os.makedirs("_build/html/doxygen", exist_ok=True)

# Inspect whether to run doxygen or not
last_modified = util.last_modified(NEK_SOURCE_ROOT).stat().st_mtime
timestamp = Path("_build/.doxygen_timestamp")
if timestamp.exists() and Path("_build/xml").exists():
    with open(timestamp) as fp:
        last_documented = float(fp.read())
    exec_doxygen = last_documented < last_modified
else:
    exec_doxygen = True

# Modify Doxygen configuration or not
#  modify_doxygen = any(os.getenv(env) for env in ("CI", "GITHUB_ACTIONS", "READTHEDOCS"))
modify_doxygen = False
if modify_doxygen:
    print("Disabling source browser... ", end="")

    # Disable source browser
    with open("Doxyfile", "rb") as doxyfile:
        doxy_cfg = [
            line for line in doxyfile.readlines() if b"SOURCE_BROWSER" not in line
        ]
    doxy_cfg = b"".join(doxy_cfg)
    # print(doxy_cfg.decode("utf8"))


try:
    if exec_doxygen:
        print("Executing Doxygen... ", end="")
        if modify_doxygen:
            # Pass configuration via stdin
            with subprocess.Popen(["doxygen", "-"], stdin=PIPE, stdout=PIPE) as proc:
                doxy_output = proc.communicate(input=doxy_cfg)[0]
        else:
            doxy_output = subprocess.check_output(["doxygen"])

        doxy_summary = doxy_output.decode("utf8").splitlines()[-2:]
        print("done:", *doxy_summary)
        with open(timestamp, "w") as fp:
            fp.write(str(last_modified))
    else:
        print(
            f"Using old Doxygen XML output... Remove {timestamp} to force doxygen build."
        )
except FileNotFoundError:
    print("Can not find doxygen to generate the documentation of the Fortran code.")
else:
    # -- Breathe configuration ---------------------------------------------------
    extensions.append("breathe")

    # File types
    breathe_implementation_filename_extensions = [".md"]
    breathe_domain_by_extension = {"usr": "fortran", "f": "fortran"}
    breathe_domain_by_file_pattern = {"SIZE": "f"}

    # Input sources
    breathe_projects = {"snek5000": "_build/xml/"}
    breathe_default_project = "snek5000"

# ----------------------------------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
#  source_suffix = [".rst", ".md"]


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "examples/snek5000-*/README.md",
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"


# Set link name generated in the top bar.
html_title = ""

html_theme_options = {
    k: v
    for k, v in runpy.run_path("html_theme_options.py").items()
    if not k.startswith("__")
}
#  html_sidebars = {
#      "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
#  }

html_favicon = "_static/favicon.ico"
html_logo = "_static/icon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Values to pass into the template engine's context for all pages.
html_context = {
    "sidebar_external_links_caption": "Links",
    "sidebar_external_links": [
        (
            '<i class="fa fa-cube fa-fw"></i> PyPI',
            f"https://pypi.org/project/{project.lower()}",
        ),
        #  (
        #      '<i class="fa fa-cube fa-fw"></i> Conda forge',
        #      f"https://anaconda.org/conda-forge/{project.lower()}",
        #  ),
        (
            '<i class="fa fa-code fa-fw"></i> Source code',
            f"https://github.com/snek5000/{project.lower()}",
        ),
        (
            '<i class="fa fa-bug fa-fw"></i> Issue tracker',
            f"https://github.com/snek5000/{project.lower()}/issues",
        ),
        #  ('<i class="fa fa-rss fa-fw"></i> Blog', 'https://...'),
        #  (
        #      '<i class="fa fa-comments fa-fw"></i> Chat',
        #      "https://matrix.to/#/#snek5000:matrix.org",
        #  ),
        #  (
        #      '<i class="fa fa-file-text fa-fw"></i> Citation',
        #      "https://doi.org/10.5334/jors.237",
        #  ),
    ],
}
# -- Options for Intersphinx -------------------------------------------------
intersphinx_mapping = runpy.run_path("ls_intersphinx_targets.py")["intersphinx_mapping"]

# -- Other options ------------------------------------------------------------
autosummary_generate = True

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    #  'special-members': '__init__',
    "undoc-members": True,
    "exclude-members": "__weakref__",
    "inherited-members": True,
}
autodoc_mock_imports = ["IPython"]

todo_include_todos = True

napoleon_numpy_docstring = True

# -- Custom functions --------------------------------------------------------


def autodoc_skip_member(app, what, name, obj, skip, options):
    # return True if (skip or exclude) else None
    # Can interfere with subsequent skip functions.
    if what == "function" and name == "load" and obj is snek5000.load:
        return True


def setup(app):
    app.connect("autodoc-skip-member", autodoc_skip_member)


myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]
myst_heading_anchors = 3



"""List files with no extensions"""
import os
from pathlib import Path

try:
    NEK_SOURCE_ROOT = Path(os.environ["NEK_SOURCE_ROOT"])
except KeyError:
    raise RuntimeError("NEK_SOURCE_ROOT has to be defined")


def is_upper(name):
    return name.upper() == name


def ls_no_ext(root, suffix="", prefix=""):
    files = Path(root).glob("*")
    for file in files:
        if not file.suffix and is_upper(file.name):
            print("{}{}{}".format(suffix, file.name, prefix))


if __name__ == "__main__":
    # For doxygen
    ls_no_ext(NEK_SOURCE_ROOT / "core", " " * 25, " \\")
    ls_no_ext("../src/abl/toolbox", " " * 25, " \\")

    # For vim
    # ls_no_ext(NEK_SOURCE_ROOT / "core", " "*12 + "\\ ", ",")



"""Load and parse spatial means data."""

from pathlib import Path

import pandas as pd


class SpatialMeans:
    INDEX_USERPARAM_SPATIAL_MEANS = 11
    _tag = "spatial_means"

    def __init__(self, output=None) -> None:
        self.output = output
        self.path_file = Path(output.path_run) / (self._tag + ".csv")

    @classmethod
    def _complete_params_with_default(cls, params):
        params.output._set_child(
            cls._tag,
            attribs={"write_interval": 1.0},
        )
        params.output.spatial_means._record_nek_user_params(
            {"write_interval": cls.INDEX_USERPARAM_SPATIAL_MEANS}
        )

    def load(self):
        return pd.read_csv(self.path_file)

    def plot(self, **kwargs):
        df = self.load()
        ax = df.plot("time", **kwargs)
        ax.figure.tight_layout()



"""Read field files using ``pymech`` as ``xarray`` datasets."""

import pymech as pm

from ...log import logger
from . import ReaderBase


class ReaderPymech(ReaderBase):
    tag = "pymech"

    def load(self, prefix="", index=-1, t_approx=None, **kwargs):
        """Opens field files(s) as a xarray dataset. The data is cached in
        :attr:`data`.

        Parameters
        ----------
        prefix: str
            Field file prefix to load custom output files. Empty for default
            field files.

        index : int or str
            When an integer is provided, it refers to the index position in a
            sorted list of field files. When ``index == "all"`` all files are
            loaded. When ``index`` is some other string, it is assumed to be
            glob pattern for the final 5 digits of the field filename extension
            (for example: ``"??20?"``).

        t_approx: float
            Find a file from approximate simulation time

        **kwargs
            Keyword arguments for ``pymech.open_*`` function

        Returns
        -------
        ds: xarray.Dataset

        """
        if isinstance(index, int):
            path_file = self.output.get_field_file(prefix, index, t_approx)
            ds = pm.open_dataset(path_file, **kwargs)
        elif isinstance(index, str):
            case = self.output.name_solver
            path = self.output.path_session
            ext = "?????" if index == "all" else index

            pattern = f"{prefix}{case}0.f{ext}"
            ds = pm.open_mfdataset(path.glob(pattern), **kwargs)
        else:
            raise ValueError("Parameter index should be int or str")

        self.data = ds
        return ds

    def get_var(self, key):
        """Return a specific array.

        Parameters
        ----------
        key: str
            Key indicating a DataArray in the loaded dataset stored in :attr:`data`.
            Must be called after :meth:`load`.

        Returns
        -------
        xarray.DataArray

        """
        if not self.data:
            logger.info("Using defaults of the load() method to read the data.")
            self.load()

        return self.data[key]


class ReaderPymechStats(ReaderPymech):
    tag = "pymech_stats"

    def load(self, prefix="sts", index=-1, **kwargs):
        """Opens statistics_ field files(s) as a xarray dataset

        .. _statistics: https://kth-nek5000.github.io/KTH_Framework/group__stat.html

        """
        return super().load(prefix, index, **kwargs)



import sys

# NOTE: If intersphinx_mapping is updated, add the keys to the Makefile
# variable INTERSPHINX_TARGETS

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    nek=("https://nek5000.github.io/NekDoc", None),
    jinja2=("https://jinja.palletsprojects.com/en/3.0.x", None),
    fluiddyn=("https://fluiddyn.readthedocs.io/en/latest", None),
    fluidsim=("https://fluidsim.readthedocs.io/en/latest", None),
    pymech=("https://pymech.readthedocs.io/en/latest", None),
)


if __name__ == "__main__":
    try:
        url = "/".join((intersphinx_mapping[sys.argv[1]][0], "objects.inv"))
    except (KeyError, IndexError):
        print(
            f"""
usage: {sys.argv[0]} [PROJECT]

POSITIONAL ARGUMENTS
  project \t one of the following {intersphinx_mapping.keys()}
    """
        )
        raise

    from sphinx.ext.intersphinx import inspect_main

    inspect_main([url])



"""Miscellaneous utilities
==========================

.. autosummary::
   :toctree:

   archive
   console
   files
   restart
   smake

"""
import itertools
import os
import sys
from datetime import datetime
from functools import partial, reduce
from tarfile import TarFile
from zipfile import ZipFile

from fluiddyn import util

from .. import get_nek_source_root
from .restart import get_status

__all__ = ["get_nek_source_root", "get_status"]

repeat = partial(itertools.repeat, None)
repeat.__doc__ = """\
Iterator which returns a ``None`` object for a number of times.

.. note::

        Using :func:`itertools.repeat` instead of :func:`range` is considered a
        more efficient option when the iterated value is unused.
        https://stackoverflow.com/a/9098860


Examples
--------
>>> for _ in repeat(10):
...     # do something


"""


def isoformat(dt):
    """Returns timestamp in modified isoformat from a datetime object (-
    instead of :). Accurate upto seconds.

    .. note:: The modified ISO format is YYYY-MM-DDTHH-MM-SS

    """
    return dt.isoformat(timespec="seconds").replace(":", "-")


def timestamp(path):
    """Modification date of a file or directory as a timestamp.

    :returns str:

    """
    return isoformat(util.modification_date(path))


def modification_date(path):
    """Modification date of a file or directory. Returns empty string if does
    not exist.

    """
    if os.path.exists(path):
        t = os.path.getmtime(path)
        return isoformat(datetime.fromtimestamp(t))
    else:
        return ""


def now():
    """The current timestamp.

    :returns str:

    """
    return isoformat(datetime.now())


def zip_info(filename):
    """Contents of a zip file."""
    with ZipFile(filename) as file:
        # Print all contents
        file.printdir()


def tar_info(filename):
    """Contents of a tar file."""
    with TarFile(filename) as file:
        # Print all contents
        file.list()


def scantree(path):
    """Recursively yield DirEntry objects for given directory.

    Courtesy: https://stackoverflow.com/a/33135143

    :returns generator: A generator for ``DirEntry`` objects.

    """
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            yield from scantree(entry.path)
        else:
            yield entry


def last_modified(path):
    """Find the last modified file in a directory tree.

    :returns DirEntry:

    """
    return reduce(
        lambda x, y: x if x.stat().st_mtime > y.stat().st_mtime else y,
        scantree(path),
    )


def init_params(Class, isolated_unit=False):
    """Instantiate an isolated ``params`` for a specific class."""
    if hasattr(Class, "create_default_params") and not isolated_unit:
        params = Class.create_default_params()
    else:
        from ..params import Parameters

        params = Parameters(tag="params")
        try:
            Class._complete_params_with_default(params)
        except TypeError:
            from snek5000.info import InfoSolverMake

            info_solver = InfoSolverMake()

            Class._complete_params_with_default(params, info_solver)

    return params


def docstring_params(Class, sections=False, indent_len=4):
    """Creates a parameter instance and generate formatted docs for it. The
    docs are defined by the ``params.<child>._set_doc`` method. Done only when
    ``sphinx`` is already imported.

    """
    if "sphinx" in sys.modules:
        params = init_params(Class, isolated_unit=True)
        doc = "\n" + params._get_formatted_docs()
    else:
        doc = ""

    # Remove the first heading: typically no content
    doc = doc.replace("Documentation for params\n========================", "")

    if not sections:
        lines = []
        indent = " " * indent_len
        prev_line = ""

        for line in doc.splitlines():
            if line.startswith("Documentation for"):
                heading = line.lstrip()
                idx = heading.index(" for ") + 5
                # Add target for cross referencing
                lines.append(f"{indent}.. _{heading[idx:]}:\n")
                lines.append(f"{indent}**{heading}**")
            elif "Documentation for" in prev_line and any(
                line.startswith(underline) for underline in ("====", "----", "~~~~")
            ):
                continue
            else:
                lines.append(indent + line)

            prev_line = line

        doc = "\n".join((lines))

    return doc



"""Nek5000 field file readers (backend for :mod:`snek5000.output.phys_fields`)

.. autosummary::
   :toctree:

    pymech_

.. paraview_

"""
from abc import ABC, abstractmethod


class ReaderBase(ABC):
    """Abstract base class for defining the interface of all reader classes"""

    tag = NotImplemented

    @classmethod
    def _complete_params_with_default(cls, params):
        params.output.phys_fields.available_readers.append(cls.tag)

    def __init__(self, output):
        self.output = output
        self.data = None

    @abstractmethod
    def load(self, prefix="", index=-1):
        """Opens field file and loads into memory as :attr:`data`"""
        ...

    @abstractmethod
    def get_var(self, key):
        """Get an array"""
        ...



"""Nek5000 base solver
======================

A bare Nek5000 solver which does not rely on any user parameters.

"""
import math
import textwrap
from pathlib import Path

from inflection import underscore

from fluidsim_core.solver import SimulCore

from .. import logger, mpi
from ..info import InfoSolverNek
from ..params import (
    Parameters,
    _complete_params_from_xml_file,
    complete_params_from_par_file,
)
from ..util import docstring_params


class SimulNek(SimulCore):
    """Simulation class

    Parameters
    ----------

    params: Parameters
        Input parameters for the simulation run.
    write_files: bool, optional
        Write simulation files, including usr, box, SIZE files. By default
        all files are written.

    Examples
    --------

    >>> from snek5000.solvers.base import Simul
    >>> params = Simul.create_default_params()
    >>> sim = Simul(params)

    """

    InfoSolver = InfoSolverNek
    Parameters = Parameters

    @classmethod
    def create_default_params(cls):
        """Generate default parameters. ``params.nek`` contains runtime
        parameters consumed by Nek5000.

        """
        return super().create_default_params()

    @classmethod
    def load_params_from_file(cls, path_xml=None, path_par=None):
        """Instantiate a Parameters instance and populate it from
        `params_simul.xml` or `case.par`

        """
        if not (path_xml or path_par):
            raise IOError(
                "Either path to params_simul.xml or case.par should be provided"
            )

        if path_xml:
            params = Parameters(tag="params")
            _complete_params_from_xml_file(params, path_xml)
        else:
            logger.warn(
                "Loading from a par file will not have full details of the simulation"
            )
            params = cls.create_default_params()
            complete_params_from_par_file(params, path_par)

        cls._set_internal_sections(params)
        return params

    @classmethod
    def _set_internal_sections(cls, params):
        """Set internal attributes to mark user sections and disable sections
        following :attr:`InfoSolverNek.par_sections_disabled`. The internal
        attributes  ``_user`` and ``_enabled`` of :class:`Parameters` are
        modified here.

        """
        try:
            info_solver = cls.info_solver
        except AttributeError:
            info_solver = cls.InfoSolver()

        for section in info_solver.par_sections:
            getattr(params.nek, section)._set_internal_attr("_user", False)

        for section in info_solver.par_sections_disabled:
            getattr(params.nek, section)._set_internal_attr("_enabled", False)

    @classmethod
    def _complete_params_with_default(cls, params):
        """A static method used to complete the *params* container.

        The sections are:

        * ``general`` (mandatory)
        * ``problemtype``
        * ``mesh``
        * ``velocity``
        * ``pressure`` (required for velocity)
        * ``temperature``
        * ``scalar%%``
        * ``cvode``

        When scalars are used, the keys of each scalar are defined under the section
        ``scalar%%`` varying between ``scalar01`` and ``scalar99``.

        """

        params._set_child("nek")
        params_nek = params.nek

        params_nek._set_doc(
            textwrap.dedent(
                """
    The parameters in ``params.nek`` are used by Snek to produce the Nek file
    .par, which is documented here :ref:`nek:case_files_par`.

    .. note::

        For these parameters, there is nearly a direct correspondance between
        Nek5000 par file keys and Snek5000 parameter variable names, with only
        *camelCase* <-> *snake_case* conversions. This is
        implemented in :mod:`snek5000.params`.

    The sections are:

    * ``general`` (mandatory)
    * ``problemtype``
    * ``mesh``
    * ``velocity``
    * ``pressure`` (required for velocity)
    * ``temperature``
    * ``scalar%%``
    * ``cvode``

    When scalars are used, the keys of each scalar are defined under the section
    ``scalar%%`` varying between ``scalar01`` and ``scalar99``.

    There is a mechanism to enable/disable these sections so that they are used
    or not to produce the .par file (see our `tutorial on writting new solvers
    <https://snek5000.readthedocs.io/en/latest/packaging.html>`__).
"""
            )
        )

        params._set_attribs(dict(NEW_DIR_RESULTS=True, short_name_type_run="run"))

        # Referenced using Sphinx extension intersphinx. Run make intersphinx-nek in docs to see available labels
        table_in_nek_doc = {
            "general": "tab:generalparams",
            "problemtype": "tab:probtypeparams",
            "velocity": "tab:velocityparams",
            "pressure": "tab:pressureparams",
            "mesh": "tab:meshparams",
            "temperature": "tab:temperatureparams",
            "scalar01": "tab:scalarparams",
            "cvode": "tab:cvodeparams",
        }

        for section in (
            "general",
            "problemtype",
            "velocity",
            "pressure",
            "mesh",
            "temperature",
            "scalar01",
            "cvode",
        ):
            child = params_nek._set_child(section)

            child._set_doc(
                f"""
    See table :ref:`nek:{table_in_nek_doc[section]}`
"""
            )

        cls._set_internal_sections(params)

        params_nek.general._set_attribs(
            dict(
                start_from="",
                stop_at="numSteps",
                end_time=math.nan,
                num_steps=1,
                dt=math.nan,
                variable_dt=True,
                target_cfl=0.5,
                write_control="timeStep",
                write_interval=10,
                filtering=None,
                filter_modes=2,
                filter_cutoff_ratio=0.65,
                filter_weight=12.0,
                write_double_precision=True,
                dealiasing=True,
                time_stepper="BDF2",
                extrapolation="standard",
                opt_level=2,
                log_level=2,
            )
        )

        params_nek.problemtype._set_attribs(
            dict(
                equation="incompNS",
                variable_properties=False,
                stress_formulation=False,
            )
        )
        common = dict(
            residual_tol=math.nan, residual_proj=False, write_to_field_file=True
        )
        params_nek.velocity._set_attribs(common)
        params_nek.pressure._set_attribs(common)
        params_nek.temperature._set_attribs(common)
        params_nek.scalar01._set_attribs(common)

        common_ts = dict(solver="helm", advection=True, absolute_tol=math.nan)
        params_nek.temperature._set_attribs(common_ts)
        params_nek.scalar01._set_attribs(common_ts)

        params_nek.mesh._set_attrib("write_to_field_file", True)
        params_nek.velocity._set_attribs(dict(viscosity=math.nan, density=math.nan))
        params_nek.pressure._set_attrib("preconditioner", "semg_xxt")
        params_nek.temperature._set_attribs(
            dict(
                conjugate_heat_transfer=False,
                conductivity=math.nan,
                rho_cp=math.nan,
            )
        )
        params_nek.scalar01._set_attribs(dict(density=math.nan, diffusivity=math.nan))

        # Document all params
        for child_name in params.nek._tag_children:
            child = getattr(params.nek, child_name)
            child._autodoc_par(indent=4)

        return params

    def __init__(self, params):
        super().__init__(params)

        self._objects_to_print = "{:28s}{}\n".format("sim: ", type(self))
        dict_classes = self.info_solver.import_classes()

        # initialize objects
        for cls_name, Class in dict_classes.items():
            # only initialize if Class is not the Simul class
            if not isinstance(self, Class):
                obj_name = underscore(cls_name)
                setattr(self, obj_name, Class(self))
                self._objects_to_print += "{:28s}{}\n".format(
                    f"sim.{obj_name}: ", Class
                )

        if "Output" in dict_classes:
            if not params.NEW_DIR_RESULTS:
                self.path_run = self.output.path_run = Path(params.path_run)
            else:
                # path_run would be initialized by the Output instance if available
                # See self.output._init_name_run()
                self.path_run = Path(self.output.path_run)

            self.output.post_init()
        else:
            self.path_run = None
            if mpi.rank == 0:
                logger.warning("No output class initialized!")

    def create_symlink_start_from_file(self, path):
        """Create a symlink towards the start_from file"""
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"{path} does not exist.")
        path_session = Path(self.output.path_session)
        path_session.mkdir(exist_ok=True)
        dest = path_session / self.params.nek.general.start_from
        if dest.exists():
            raise FileExistsError(f"{dest} already exists.")
        logger.info(f"Symlinking {path} -> {dest}")
        dest.symlink_to(path)


Simul = SimulNek
Simul.__doc__ += """

    Notes
    -----

    Here, only the documention for ``params.nek`` is displayed. For the
    documentation on ``params.oper`` see :mod:`snek5000.operators`.

""" + docstring_params(
    Simul, indent_len=4
)



import os
import shutil
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pymech
import pytest

from snek5000 import load
from snek5000.util.gfortran_log import log_matches


def pytest_addoption(parser):
    # https://pytest.readthedocs.io/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")

    try:
        import rich  # noqa

        import snek5000.log

    except ImportError:
        pass
    else:
        # Bug while using rich + pytest: stderr / stdout is too short
        # Inspired from: https://github.com/willmcgugan/rich/issues/1425
        snek5000.log.logger.removeHandler(snek5000.log.handler)

        handler = snek5000.log.create_handler(width=shutil.get_terminal_size().columns)
        snek5000.log.logger.addHandler(handler)


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(scope="session")
def jinja_env():
    import jinja2

    env = jinja2.Environment(
        loader=jinja2.PackageLoader("snek5000", "resources"),
        undefined=jinja2.StrictUndefined,
    )
    return env


@pytest.fixture(scope="session")
def sim():
    from phill.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "test_snek5000"

    params.nek.general.stop_at = "numSteps"
    params.nek.general.num_steps = 9
    return Simul(params)


@pytest.fixture(scope="session")
def oper():
    from snek5000.operators import Operators as Class
    from snek5000.util import init_params

    params = init_params(Class)
    params.oper.nx = params.oper.ny = params.oper.nz = 9
    params.oper.nproc_min = 6

    return Class(params=params)


def set_params_oper2d(params):
    params.oper.nx = params.oper.ny = 16
    params.oper.Lx = params.oper.Ly = 1
    params.oper.dim = 2
    params.oper.boundary = ["W"] * 4
    params.oper.boundary_scalars = ["t"] * 2 + ["I"] * 2

    params.oper.nproc_min = 4
    params.oper.nproc_max = 32

    elem = params.oper.elem
    elem.order = elem.order_out = 15
    elem.coef_dealiasing = 1.0 / 1.6
    elem.staggered = False

    params.oper.max.hist = 100


@pytest.fixture(scope="session")
def oper2d():
    from snek5000.operators import Operators as Class
    from snek5000.util import init_params

    params = init_params(Class)
    set_params_oper2d(params)
    return Class(params=params)


@pytest.fixture(scope="session")
def sim2d():
    from phill.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "test_snek5000"

    set_params_oper2d(params)
    return Simul(params)


@pytest.fixture(scope="module")
def sim_canonical():
    from snek5000_canonical.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "test_snek5000"

    return Simul(params)


@pytest.fixture(scope="session")
def sim_executed():
    from phill.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "test_snek5000"

    params.nek.general.stop_at = "endTime"
    params.nek.general.end_time = 10 * abs(params.nek.general.dt)
    params.nek.general.write_interval = 5

    params.oper.nproc_min = 2
    params.oper.nproc_max = 12
    params.oper.nx = params.oper.ny = params.oper.nz = 3
    params.oper.elem.order = params.oper.elem.order_out = 8

    params.nek.stat.av_step = 2
    params.nek.stat.io_step = 4

    sim = Simul(params)
    assert sim.make.exec("run_fg"), "phill simulation failed"
    return sim


@contextmanager
def unset_snek_debug():
    old_snek_debug = os.environ.pop("SNEK_DEBUG", None)
    try:
        yield
    finally:
        if old_snek_debug is not None:
            os.environ["SNEK_DEBUG"] = old_snek_debug


@pytest.fixture(scope="session")
def sim_cbox_executed():
    from snek5000_cbox.solver import Simul

    params = Simul.create_default_params()
    params.Ra_side = 1.83e08
    params.output.sub_directory = "test_snek5000"

    params.nek.general.stop_at = "numSteps"
    params.nek.general.dt = 1e-3
    params.nek.general.num_steps = 12
    params.nek.general.write_interval = 3

    params.oper.nproc_min = 2
    params.oper.nproc_max = 12
    params.oper.dim = 2
    params.oper.nx = params.oper.ny = 8

    coords = [(0.5, 0.5)]
    params.output.history_points.write_interval = 2
    params.output.history_points.coords = coords
    params.oper.max.hist = len(coords) + 1

    sim = Simul(params)
    sim.output.write_snakemake_config(custom_env_vars={"FOO": 1})

    with unset_snek_debug():
        if not sim.make.exec("compile"):
            build_log = Path(sim.output.path_run) / "build.log"
            log_matches(build_log, levels=["Error"])
            raise RuntimeError("cbox compilation failed")

    if not sim.make.exec("run_fg", nproc=2):
        with open(Path(sim.output.path_run) / "cbox.log") as file:
            print(file.read())
        raise RuntimeError("cbox simulation failed")
    return sim


@pytest.fixture(scope="session")
def sim_cbox_executed_readonly(sim_cbox_executed):
    path_run = Path(sim_cbox_executed.output.path_run)
    with tempfile.TemporaryDirectory(suffix="snek5000_cbox_executed") as tmp_dir:
        path_readonly_sim = Path(tmp_dir) / path_run.name
        shutil.copytree(path_run, path_readonly_sim)

        for path in path_readonly_sim.rglob("*"):
            mod = 0o444
            if path.is_dir():
                mod = 0o544
            path.chmod(mod)

        sim = load(path_readonly_sim)
        yield sim

        for path in path_readonly_sim.rglob("*"):
            mod = 0o644
            if path.is_dir():
                mod = 0o744
            path.chmod(mod)


@pytest.fixture(scope="session")
def sim_tgv_executed():
    from snek5000_tgv.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "examples_snek/tuto"

    params.oper.nx = params.oper.ny = params.oper.nz = 6
    params.oper.elem.order = params.oper.elem.order_out = 6
    params.oper.nproc_min = 2

    params.nek.velocity.residual_tol = 1e-07
    params.nek.pressure.residual_tol = 1e-05

    params.nek.general.end_time = 1
    params.nek.general.dt = -1
    params.nek.general.target_cfl = 1.4
    params.nek.general.extrapolation = "OIFS"

    params.nek.general.write_control = "runTime"
    params.nek.general.write_interval = 0.5

    params.output.spatial_means.write_interval = 0.5

    sim = Simul(params)
    if not sim.make.exec("run_fg", nproc=2):
        with open(Path(sim.output.path_run) / "cbox.log") as file:
            print(file.read())
        raise RuntimeError("tgv simulation failed")

    return sim


def create_fake_nek_files(path_dir, name_solver, nb_files=1):

    nx = 2
    ny = 4
    nz = 6
    nx_elem = ny_elem = nz_elem = 2

    hexa_data = pymech.core.HexaData(
        ndim=3,
        nel=nx_elem * ny_elem * nz_elem,
        lr1=(nx, ny, nz),
        var=(3, 3, 1, 1, 0),
    )
    hexa_data.wdsz = 8
    hexa_data.istep = 0
    hexa_data.endian = sys.byteorder

    x1d = np.linspace(0, 1, nx)
    y1d = np.linspace(0, 1, ny)
    z1d = np.linspace(0, 1, nz)

    y3d, z3d, x3d = np.meshgrid(y1d, z1d, x1d)
    assert y3d.shape == (nz, ny, nx)

    ielem = 0
    for iz_elem in range(nz_elem):
        for iy_elem in range(ny_elem):
            for ix_elem in range(nx_elem):
                elem = hexa_data.elem[ielem]
                ielem += 1
                elem.pos[0] = x3d + ix_elem
                elem.pos[1] = y3d + iy_elem
                elem.pos[2] = z3d + iz_elem
                elem.vel.fill(1)

    time = 2.0
    for it in range(nb_files):
        hexa_data.time = time
        pymech.writenek(path_dir / f"{name_solver}0.f{it:05d}", hexa_data)
        time += 0.5


@pytest.fixture(scope="function")
def sim_data(tmpdir_factory):
    """Generate fake simulation data."""
    files = """box.tmp
makefile
makefile_usr.inc
nek5000
phill.box
phill.f
phill.log
phill.ma2
phill.par
phill.re2
phill.usr
rs6phill0.f00001
rs6phill0.f00002
rs6phill0.f00003
SESSION.NAME
SIZE
Snakefile""".splitlines()

    session_files = """
phill0.f00001
""".splitlines()

    path_run = Path(tmpdir_factory.mktemp("phill_sim_data"))

    from snek5000.output import _make_path_session

    for f in files:
        (path_run / f).touch()

    for session_id in range(2):
        path_session = _make_path_session(path_run, session_id)
        path_session.mkdir()
        for f in session_files:
            create_fake_nek_files(path_session, "phill")

    from phill.solver import Simul

    info = Simul.InfoSolver()
    params = Simul.create_default_params()

    info._save_as_xml(path_run / "info_solver.xml")
    params._set_attrib("path_run", path_run)
    params.output._set_attrib("session_id", session_id)
    params.output._set_attrib("path_session", path_session)
    params._save_as_xml(path_run / "params_simul.xml")

    return path_run


@pytest.fixture(autouse=True, scope="session")
def shared_datadir_remove():
    """Removes empty data directory as a result of pytest-datadir"""
    yield  # Execute all tests
    shared_datadir = Path(__file__).parent.parent / "data"
    if shared_datadir.exists():
        shared_datadir.rmdir()



"""Console script functions

"""

from textwrap import dedent

from snek5000.solvers import available_solvers


def print_versions():

    import pymech
    import snakemake

    import fluiddyn
    import fluidsim_core
    import snek5000

    versions = {"Package": "Version", "-------": "-------"}

    packages = [snek5000, fluiddyn, fluidsim_core, pymech, snakemake]
    for package in packages:
        versions[package.__name__] = package.__version__

    for pkg_name, version in versions.items():
        print(f"{pkg_name.ljust(15)} {version}")

    names = sorted(set([entry_point.name for entry_point in available_solvers()]))
    print("\nInstalled solvers: " + ", ".join(names))


def start_ipython_load_sim():
    """Start IPython and load a simulation"""
    from IPython import start_ipython

    argv = ["--matplotlib", "-i", "-c"]
    code = dedent(
        """
        import numpy as np
        import matplotlib.pyplot as plt
        import pandas as pd
        from snek5000 import load
        print("Loading simulation")
        sim = load()
        params = sim.params
        print("`sim`, `params`, `np`, `plt` and `pd` variables are available")
    """
    )
    argv.append("; ".join(code.strip().split("\n")))
    start_ipython(argv=argv)



from snek5000.util.console import print_versions, start_ipython_load_sim


def test_info():
    print_versions()


def test_start_ipython_load_sim(mocker):
    mocker.patch("IPython.start_ipython")
    start_ipython_load_sim()



from snek5000.output.base import Output as OutputBase
from snek5000.resources import get_base_templates

box, makefile_usr, size = get_base_templates()


class OutputCanonical(OutputBase):

    template_box = box
    template_makefile_usr = makefile_usr
    template_size = size

    @classmethod
    def _set_info_solver_classes(cls, classes):
        """Set the the classes for info_solver.classes.Output

        NOTE: This method is optional, and only required if custom classes are
        to be set.

        """
        super()._set_info_solver_classes(classes)

        # Replace sim.output.phys_fields with a custom class
        classes.PhysFields.module_name = "snek5000_canonical.phys_fields"
        classes.PhysFields.class_name = "PhysFieldsCanonical"

    @property
    def makefile_usr_sources(self):
        """
        Sources for inclusion to makefile_usr.inc
        Dict[directory]  -> list of source files
        """
        return {}

        # For example this is the list of extra files required for the KTH
        # Framework:

        # return {
        #     "toolbox": [
        #         ("frame.f", "FRAMELP"),
        #         ("mntrlog_block.f", "MNTRLOGD"),
        #         ("mntrlog.f", "MNTRLOGD"),
        #         ("mntrtmr_block.f", "MNTRLOGD", "MNTRTMRD"),
        #         ("mntrtmr.f", "MNTRLOGD", "MNTRTMRD", "FRAMELP"),
        #         ("rprm_block.f", "RPRMD"),
        #         ("rprm.f", "RPRMD", "FRAMELP"),
        #         ("io_tools_block.f", "IOTOOLD"),
        #         ("io_tools.f", "IOTOOLD"),
        #         ("chkpoint.f", "CHKPOINTD"),
        #         ("chkpt_mstp.f", "CHKPTMSTPD", "CHKPOINTD"),
        #         ("map2D.f", "MAP2D", "FRAMELP"),
        #         ("stat.f", "STATD", "MAP2D", "FRAMELP"),
        #         ("stat_IO.f", "STATD", "MAP2D", "FRAMELP"),
        #         ("math_tools.f",),
        #     ],
        # }


Output = OutputCanonical



"""
Solver framework

.. autosummary::
   :toctree:

   base
   kth

"""
import importlib
from functools import partial
from pkgutil import ModuleInfo
from types import ModuleType

from fluidsim_core import loader

from ..log import logger

available_solvers = partial(loader.available_solvers, entrypoint_grp="snek5000.solvers")
available_solvers.__doc__ = """\
Returns a dictionary of all registered solvers registered as an entrypoint.
"""

import_cls_simul = partial(loader.import_cls_simul, entrypoint_grp="snek5000.solvers")
import_cls_simul.__doc__ = """Import the Simul class of a solver."""


def is_package(module):
    """Checks if a module is a package or not. As of PEP 451 this information
    is also available on the module’s __spec__.submodule_search_locations
    attribute, which will not be None for packages.

    Parameters
    ----------
    module: str or module

    Returns
    -------
    bool

    """
    if isinstance(module, str):
        spec = importlib.util.find_spec(module)
    elif isinstance(module, ModuleType):
        spec = module.__spec__
    elif isinstance(module, ModuleInfo):
        return module.ispkg
    else:
        raise ValueError(f"Cannot process {type(module)}")

    return spec.submodule_search_locations is not None


def get_solver_package(name_solver):
    """Return name of the solver package by checking the ``snek5000.solvers``
    entrypoint group.

    Parameters
    ----------
    name_solver: str
        Short name of the solver

    Returns
    -------
    str

    """
    entrypoint = available_solvers()[name_solver]
    try:
        # importlib.metadata.entry_points (fluidsim_core > 0.6.0)
        module = entrypoint.module
    except AttributeError:
        module = entrypoint.module_name
    if is_package(module):
        return module
    else:
        return module.rpartition(".")[0]


def get_solver_short_name(path_dir):
    """Detects short name of the solver from ``info_solver.xml`` if present or
    the path.

    Parameters
    ----------
    path_dir: path-like
        Path to a simulation directory

    Returns
    -------
    short_name: str

    """
    info_solver_xml = path_dir / "info_solver.xml"
    if info_solver_xml.exists():
        from snek5000.info import InfoSolverNek

        info_solver = InfoSolverNek(path_file=info_solver_xml)
        short_name = info_solver.short_name
    else:
        logger.warning(
            f"The {info_solver_xml} file is missing! "
            "Attempting to guess solver from the directory name."
        )
        short_name = path_dir.absolute().name.split("_")[0]

    return short_name



import os

from snek5000.util.archive import archive, clean_simul, tar_name


def test_name(sim_data):
    root_name = "phill"
    pattern = str(sim_data / "*")
    ark_tar_gz = tar_name(root_name, pattern, compress_format=".gz")
    assert ark_tar_gz


def test_compress(sim_data):
    contents = sorted(sim_data.iterdir())
    half = len(contents) // 2

    archive(sim_data / "ark.tar", contents[:half], remove=True)
    archive(sim_data / "ark.tar", contents[half:])
    archive(sim_data / "ark.tar.gz", readonly=True)

    os.chdir(str(sim_data))
    clean_simul("phill", "ark.tar.gz")



import re
from pathlib import Path

from snek5000.log import logger

pattern = r"""
(?P<file>[^:^\n]+):(?P<line_nb>\d+):(?P<col>\d+):\n{,3}              # Filename and location: file, line_nb, col
(\s*?(?P<L>\d+?)\ \|(?P<source>.*?)$\n\s*?\|(?P<mark>.*?\d)\n\.*?)*  # Source code: L, source, mark
^(?P<level>Warning|Error):\ (?P<msg>.*)                              # Reason: level, msg
"""

expr = re.compile(pattern, re.MULTILINE | re.VERBOSE)


def log_match(match, levels):
    """Display a match which is included in the levels requested.

    Parameters
    ----------
    match : re.Match
        A regular expression match
    levels : iterable of str
        Should by a subset of `{"Error", "Warning"}`

    """
    if match["level"] in levels:
        logger.info(
            f"{match['level']}: "
            f"{match['file']}:{match['line_nb']}:{match['col']}\n"
            f"{match.group(match.lastindex)}\n"
            #  match["source"],
        )


def log_matches(path_log, levels=("Error", "Warning")):
    """Log all matches

    Parameters
    ----------
    path_log : str or path-like
        Path to a log file with gfortran compilation output
    levels : iterable of str
        Should by a subset of `{"Error", "Warning"}`
    """
    text = Path(path_log).read_text()
    for match in expr.finditer(text):
        log_match(match, levels=["Error"])



import logging
import os
from pathlib import Path

import pytest

from snek5000.config import ensure_config_file
from snek5000.output.base import Output, missing_config_filter

xdg_config = Path(os.path.expandvars(os.getenv("XDG_CONFIG_HOME", "$HOME/.config")))
configfile_xdg_config = xdg_config / "snek5000.yml"


def test_update_snakemake_config():
    # Dummy config values
    config = {
        key: ""
        for key in (
            "CC",
            "FC",
            "MPICC",
            "MPIFC",
            "MPIEXEC",
            "MPIEXEC_FLAGS",
            "CFLAGS",
            "FFLAGS",
        )
    }
    name_solver = "base"
    Output.update_snakemake_config(config, name_solver)

    expected_keys = {"CASE", "file", "includes", "objects"}
    assert expected_keys.issubset(config)

    original_case = config["CASE"]
    os.environ["CASE"] = "test_config"

    Output.update_snakemake_config(config, name_solver, env_sensitive=True)
    assert original_case != "test_config"
    assert config["CASE"] == "test_config"

    os.environ["SNEK_UPDATE_CONFIG_ENV_SENSITIVE"] = "1"
    os.environ["CASE"] = "test_config_1"
    Output.update_snakemake_config(config, name_solver)
    assert config["CASE"] == "test_config_1"

    os.environ["SNEK_UPDATE_CONFIG_ENV_SENSITIVE"] = "0"
    os.environ["CASE"] = "test_config_2"
    Output.update_snakemake_config(config, name_solver)
    assert config["CASE"] == original_case


@pytest.mark.skipif(
    configfile_xdg_config.exists(), reason=f"File {configfile_xdg_config} exists"
)
def test_find_configfile(caplog):
    hostname = "test_config"

    missing_config_filter.emitted = False
    Output.find_configfile(host=hostname)

    for logger_name, level, message in caplog.record_tuples:
        if (
            logger_name == "snek5000.log"
            and level == logging.WARNING
            and "Missing a configuration file" in message
        ):
            assert f"{hostname}.yml" in message
            break
    else:
        raise IOError("Expected a warning message when configuration file is missing")


@pytest.mark.skipif(
    configfile_xdg_config.exists(), reason=f"File {configfile_xdg_config} exists"
)
def test_ensure_config_file():
    config_file = Output.find_configfile()
    ensure_config_file()
    ensure_config_file()
    if config_file.name == "default_configfile.yml":
        configfile_xdg_config.unlink()



"""Interface for Nek5000 field files.

:class:`PhysFields` usually instantiated as ``sim.output.phys_fields`` provides
a common interface for plotting and reading arrays from the solution field
files. Loading of data files are managed by the classes in the
:mod:`snek5000.output.readers`.

"""

from fluidsim_core.output.phys_fields_snek5000 import PhysFields4Snek5000
from fluidsim_core.params import iter_complete_params

from ..log import logger

#  from .readers import try_paraview_ as pv
from .readers import pymech_ as pm


class PhysFields(PhysFields4Snek5000):
    """Class for loading, plotting simulation files."""

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        """Static method to complete the ParamContainer info_solver.

        Parameters
        ----------

        info_solver : :class:`snek5000.info.InfoSolverMake`

        classes : iterable of classes

          If a class has the same tag of a default class, it replaces the
          default one (for example with the tag 'pymech').

        """
        classes_xml = info_solver.classes.PhysFields._set_child("classes")

        avail_classes = [
            #  pv.ReaderParaview,
            #  pv.ReaderParaviewStats,
            pm.ReaderPymech,
            pm.ReaderPymechStats,
        ]
        if classes is not None:
            avail_classes.extend(classes)

        for cls in avail_classes:
            classes_xml._set_child(
                cls.tag,
                attribs={
                    "module_name": cls.__module__,
                    "class_name": cls.__name__,
                },
            )

    @classmethod
    def _complete_params_with_default(cls, params, info_solver):
        params.output._set_child(
            "phys_fields", attribs={"reader": "pymech", "available_readers": []}
        )

        dict_classes = info_solver.classes.Output.classes.PhysFields.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    @property
    def data(self):
        data = self._reader.data
        if not data:
            self.load()
            data = self._reader.data
        return data

    def __init__(self, output=None):
        super().__init__(output)
        self._reader = None  #: Reader instance

        self.load = self._uninitialized
        """Reader method which loads a particular file into memory and returns it.

        .. seealso::
            :meth:`snek5000.output.readers.ReaderBase.load`
        """

        self.get_var = self._uninitialized
        """Reader method which returns a particular array from memory.

        .. seealso::
            :meth:`snek5000.output.readers.ReaderBase.get_var`
        """

    def _uninitialized(self, *args, **kwargs):
        """Place holder method to raise a :exc:`RuntimeError` while accessing
        :meth:`init_reader` not initialized.

        """
        raise RuntimeError(
            "The reader and the method has not initialized yet. "
            "Call sim.output.phys_fields.init_reader() to initialize the reader."
        )

    def init_reader(self):
        """Initializes the reader instance following the value in
        ``params.output.phys_fields.reader``. This would also "initialize"
        :meth:`load` and :meth:`get_var`.

        """
        if self._reader:
            logger.warning(
                "The reader is already initialized. Use change_reader() "
                "if you need to change the backend"
            )
            return

        reader_key = self.params.phys_fields.reader
        self.change_reader(reader_key)

    def change_reader(self, reader_key):
        """Changes the reader following which ``reader_key`` is provided.

        Parameters
        ----------
        reader_key: str
            String denoting the reader class. Should be one of
            ``params.output.phys_fields.available_readers``.
        """
        sim = self.output.sim

        dict_classes = (
            sim.info.solver.classes.Output.classes.PhysFields.import_classes()
        )
        try:
            cls = dict_classes[reader_key]
        except AttributeError as err:
            available_readers = self.params.phys_fields.available_readers
            raise ValueError(
                f"{reader_key =} not found in {available_readers =}"
            ) from err

        self._reader = cls(self.output)
        self.load = self._reader.load
        self.get_var = self._reader.get_var



import os
from pathlib import Path

from snek5000.util import files


def test_next_path_no_files(tmpdir):
    tmpdir = Path(tmpdir)

    # Create no files, multiple suffixes
    target = tmpdir / "test.tar.gz"

    assert str(files.next_path(target, force_suffix=True)) == str(
        tmpdir / "test_00.tar.gz"
    )


def test_next_path_one_file(tmpdir):
    tmpdir = Path(tmpdir)

    # Create a file
    target = tmpdir / "test.txt"
    target.touch()  # Not necessary, but to see what happens.

    assert str(files.next_path(target, force_suffix=True)) == str(
        tmpdir / "test_00.txt"
    )


def test_next_path_one_file_no_suffix(tmpdir):
    tmpdir = Path(tmpdir)

    # Create a file
    target = tmpdir / "test.txt"
    target.touch()
    (tmpdir / "test_00.txt").touch()

    assert str(files.next_path(target)) == str(tmpdir / "test_01.txt")


def test_next_path_two_files(tmpdir):
    tmpdir = Path(tmpdir)

    # Create multple files
    target = tmpdir / "test.txt"
    (tmpdir / "test_00.txt").touch()
    (tmpdir / "test_01.txt").touch()

    assert str(files.next_path(target, force_suffix=True)) == str(
        tmpdir / "test_02.txt"
    )


def test_next_path_dir_no_suffix(tmpdir):
    tmpdir = Path(tmpdir)

    # Create a directory
    target = tmpdir / "test_dir"
    target.mkdir()
    (target / "foo").touch()

    assert str(files.next_path(target)) == str(tmpdir / "test_dir_00")


def test_next_path_dir(tmpdir):
    tmpdir = Path(tmpdir)

    target = tmpdir / "test_dir"
    path_dir = files.next_path(target, force_suffix=True)

    assert str(path_dir) == str(tmpdir / "test_dir_00")


def test_next_path_relative(tmpdir):
    os.chdir(tmpdir)

    session_00 = Path("./session_00")
    session_01 = Path("./session_01")

    assert files.next_path("session", force_suffix=True) == session_00

    session_00.mkdir()
    (session_00 / "foo").touch()
    int_suffix, session_dir = files.next_path(
        "session", force_suffix=True, return_suffix=True
    )

    assert int_suffix == 1
    assert session_dir == session_01

    if session_dir.is_absolute():
        raise ValueError("next_path should return a relative path")



"""Information classes
======================

Store information regarding modules and class names of simulation class and
dependent objects.

"""
from fluidsim_core.info import InfoSolverCore


class InfoSolverNek(InfoSolverCore):
    """Contain the information on a :class:`snek5000.solvers.base.Simul`
    instance.

    """

    def _init_root(self):
        """Defines class hierarchy (via :attr:`classes`) and some attributes

        - :attr:`module_name` - module of the Simul class
        - :attr:`class_name` - Simul class name
        - :attr:`short_name` - short name of the solver (and name used in the
          source code usr, par, re2, ma2 etc.)
        - :attr:`par_sections` - sections in the .par file
        - :attr:`par_sections_disabled` - sections to be disabled (i.e. removed
          while generating) in the .par file

        """
        self._set_attribs(
            {
                "module_name": "snek5000.solvers.base",
                "class_name": "SimulNek",
                "short_name": "nek",
                "loader": "snek5000.load",
                "par_sections": (
                    "general",
                    "problemtype",
                    "velocity",
                    "pressure",
                    "mesh",
                    "temperature",
                    "scalar01",
                    "cvode",
                ),
                "par_sections_disabled": {
                    "mesh",
                    "temperature",
                    "scalar01",
                    "cvode",
                },
            }
        )
        self._set_child("classes")
        self.classes._set_child(
            "Oper",
            attribs={"module_name": "snek5000.operators", "class_name": "Operators"},
        )


class InfoSolverMake(InfoSolverNek):
    """Contain the information on a solver which uses Snakemake."""

    def _init_root(self):
        super()._init_root()
        self.classes._set_child(
            "Output",
            attribs={"module_name": "snek5000.output.base", "class_name": "Output"},
        )
        self.classes._set_child(
            "Make", attribs={"module_name": "snek5000.make", "class_name": "Make"}
        )



import sys
from concurrent.futures import ProcessPoolExecutor as Pool
from concurrent.futures import as_completed
from pathlib import Path
from unittest.mock import patch

import pytest

from snek5000.make import _Nek5000Make, snek_make


def nek5000_build(config):
    nm = _Nek5000Make()
    return nm.build(config)


def test_nek5000_make(pytestconfig):
    import yaml

    import snek5000
    from snek5000.output.base import Output

    if pytestconfig.getoption("--runslow"):
        # To force rebuild of Nek5000
        (Path(snek5000.get_nek_source_root()) / "bin" / "genbox").unlink()

    snek5000_config = yaml.safe_load(Output.find_configfile().read_text())

    with Pool(max_workers=4) as pool:
        simultaneous_builds = [
            pool.submit(nek5000_build, snek5000_config) for _ in range(4)
        ]
        for future in as_completed(simultaneous_builds):
            assert future.result()


def test_snek_make_help():
    with patch.object(sys, "argv", ["snek-make", "-h"]):
        with pytest.raises(SystemExit):
            snek_make()


def test_snek_make_no_rule(sim_data, monkeypatch):
    monkeypatch.chdir(sim_data)
    with patch.object(sys, "argv", ["snek-make"]):
        with pytest.raises(SystemExit):
            snek_make()


def test_snek_make_clean(sim_data, monkeypatch):
    monkeypatch.chdir(sim_data)
    with patch.object(sys, "argv", ["snek-make", "clean"]):
        snek_make()



from functools import cached_property
from pathlib import Path

import numpy as np

from ...log import logger
from . import ReaderBase

try:
    import paraview.simple as pv
    from paraview.simple import *  # noqa: F401, F403
except ImportError:
    logger.debug("Paraview Python module cannot be imported")
else:
    # Continue importing VTK
    from vtk.numpy_interface import dataset_adapter as dsa  # algorithms as algs,
    from vtk.util.numpy_support import vtk_to_numpy


#  import pyvista
# from vtktools import vtkio


# disable automatic camera reset on 'Show'
# pv._DisableFirstRenderCameraReset()


class ReaderParaview(ReaderBase):
    """A user-friendly API for Paraview scripting with 'VisItNek5000Reader'."""

    tag = "paraview"

    def open(
        self,
        filename,  # <case>.nek5000
        arrays=(
            "pressure",
            "velocity_mag",
            "x_velocity",
            "y_velocity",
            "z_velocity",
        ),
    ):
        self.filename = filename
        assert Path(filename).exists()

        self.nek5000 = nek5000 = pv.VisItNek5000Reader(FileName=str(filename))
        nek5000.Meshes = ["mesh"]

        # get animation scene
        self.animationScene1 = pv.GetAnimationScene()

        # get the time-keeper
        self.timeKeeper1 = pv.GetTimeKeeper()

        # update animation scene based on data timesteps
        self.animationScene1.UpdateAnimationUsingDataTimeSteps()

        # Properties modified on ablnek5000
        assert isinstance(arrays, (list, tuple))
        nek5000.PointArrays = list(arrays)

        data_info = nek5000.GetDataInformation()
        self.dtype = data_info.GetDataSetTypeAsString()

    def __iter__(self):
        return self

    def __next__(self):
        """Iterate through time"""
        if self.time == self.animationScene1.EndTime:
            raise StopIteration

        self.animationScene1.GoToNext()
        self.animationScene1.AnimationTime = self.time

        #  self.animationScene1.UpdateAnimationUsingDataTimeSteps = self.time
        #  pv.UpdatePipeline(self.time)
        self.apply()

        # Return lightweight metadata as a tuple
        return (self.time,)

    @property
    def time(self):
        return self.timeKeeper1.Time

    @time.setter
    def time(self, value):
        """Seek a new time step and update pipeline."""
        # Properties modified on animationScene1
        self.animationScene1.AnimationTime = value
        # Properties modified on timeKeeper1
        self.timeKeeper1.Time = value
        self.apply()

    @property
    def timesteps(self):
        return self.timeKeeper1.TimestepValues

    @cached_property
    def renderView1(self):
        """Get active view."""
        return pv.GetActiveViewOrCreate("RenderView")

    @cached_property
    def temporalStatistics1(self):
        temporalStatistics1 = pv.TemporalStatistics(Input=self.nek5000)

        # Properties modified on temporalStatistics1
        temporalStatistics1.ComputeAverage = 1
        temporalStatistics1.ComputeMinimum = 0
        temporalStatistics1.ComputeMaximum = 0
        return temporalStatistics1

    @cached_property
    def groupDatasets1(self):
        return pv.GroupDatasets(Input=[self.temporalStatistics1, self.nek5000])

    @cached_property
    def bounds(self, source=None):
        Input = source if source else self.nek5000
        bounds = Input.GetDataInformation().GetBounds()
        if 1.0e299 in bounds:
            self.apply()
            bounds = Input.GetDataInformation().GetBounds()
        return bounds

    def show(self, key_array="x_velocity", source=None, rescale=False):
        """Show the rendered output on the screen."""
        Input = source if source else self.nek5000

        display = pv.Show(Input, self.renderView1)
        pv.ColorBy(display, ("POINTS", key_array))
        if rescale:
            display.RescaleTransferFunctionToDataRange(True, False)
        display.SetScalarBarVisibility(source, True)

        # ... lot of parameters are possible here

        pv.SetActiveSource(source)

        try:
            LUT = pv.GetColorTransferFunction(key_array)
            PWF = pv.GetOpacityTransferFunction(key_array)
        except AttributeError:
            LUT = PWF = None

        return display, LUT, PWF

    def apply(self):
        """Update pipeline with current timestep."""
        pv.UpdatePipeline(self.time)

    def render(self):
        renderView1 = self.renderView1
        renderView1.ResetCamera()
        renderView1.Update()

    def get_slice(self, x=0.0, y=0.0, z=0.0, normal=(0, 1, 0), source=None):
        slice1 = self.slice1 = pv.Slice(Input=source if source else self.nek5000)
        slice1.SliceType = "Plane"
        slice1.SliceOffsetValues = [0.0]
        slice1.SliceType.Origin = [x, y, z]
        slice1.SliceType.Normal = list(normal)
        # slice1.Triangulatetheslice = 0

        # slice_data = vtkio.getBlockByName(mb_dset, 'mesh')

        # pv.SetActiveSource(slice1)

        return _ParaviewSlice(slice1)
        # return pyvista.wrap(dobj)


class ReaderParaviewStats(ReaderParaview):
    tag = "paraview_stats"

    def open(
        self,
        filename,  # "sts<case>.nek5000",
        arrays=tuple(f"s{idx}" for idx in range(1, 45)),
    ):
        super().__init__(filename, arrays)

    def calculate(self, name="u_prime", func="s1_average - s1"):
        calculator1 = pv.Calculator(Input=self.groupDatasets1)
        calculator1.ResultArrayName = name
        calculator1.Function = func
        return calculator1


class _ParaviewSlice:
    """Convenience wrapper to easily access arrays stored in a VTK dataset."""

    def __init__(self, slice1):
        self._slice = slice1
        # vtkCommonDataModelPython.vtkMultiBlockDataSet
        dset = pv.servermanager.Fetch(slice1)

        if dset.GetClassName() not in ("vtkMultiBlockDataSet",):
            raise ValueError(f"Incompatible type: {type(dset)}")

        # <vtk.numpy_interface.dataset_adapter.CompositeDataSet
        self._dset = dset
        self._obj = dsa.WrapDataObject(dset)
        self._ptdata = self._obj.GetPointData()

    def __getitem__(self, key):
        keys = self.keys()

        if key in self._ptdata.keys():
            return self._ptdata.GetArray(key)
        elif key == "coords":
            return self._obj.Points
        else:
            raise KeyError(f"Valid keys are: {keys}")

    def keys(self):
        return ["coords"] + self._ptdata.keys()

    def get_blocks(self):
        nb_blocks = self._dset.GetNumberOfBlocks()
        return (self._dset.GetBlock(idx) for idx in range(nb_blocks))

    def get_array(self, key):
        vtk_array = self[key].Arrays[0]
        array = vtk_to_numpy(vtk_array)
        return array

    def get_coords(self, normal=1, sort=True, reshape=True):
        """Get row major sorted and reshaped coordinates for a slice.

        :param int normal: Axis normal to the slice
        :param bool sort: Sort the coordinate array or not
        :param bool reshape: Reshape the coordinates into 2D array or not.

        """
        coords = self.get_array("coords")
        # Remove normal directions
        axes = {0, 1, 2} - {normal}
        ax0, ax1 = sorted(axes)

        def sorting_indices(pts):
            """Sorts indices using a weighting function."""
            pts = np.copy(pts)

            # weighting_func = lambda x0, x1: x0 + x1 * 1e4
            # inds = np.argsort(weighting_func(pts[:, ax0], pts[:, ax1]))
            inds = np.lexsort([pts[:, ax] for ax in range(3)])
            return inds

        if sort:
            inds = sorting_indices(coords)
            coords = coords[inds]
        else:
            inds = None

        # Sorting coordinates
        def get_n0_n1(pts):
            """Estimate array shape"""
            # Look where index jumps occur
            # ----------------------------
            # For example 0, 1, 2, 3, (jump!) 0, 1, 2, 3
            inds_jumps = np.where(pts[:-1, ax0] > pts[1:, ax0])[0]
            # print(inds_jumps)
            n1 = inds_jumps[1] - inds_jumps[0]
            n1_avg = np.diff(inds_jumps).mean()
            assert (
                n1 == n1_avg
            ), f"{n1}!={n1_avg}, i.e. inhomogenieties found, reshaping may not be possible"

            n1 = int(n1)

            n0 = pts.shape[0] / n1
            assert n0.is_integer()
            n0 = int(n0)

            return n0 + 1, n1

        x, y, z = coords.T
        if reshape:
            n0, n1 = get_n0_n1(coords)
            axes = [self.reshape(ax, shape=(n0, n1)) for ax in (x, y, z)]
            return (*axes, inds)
        else:
            return x, y, z, inds

    def reshape(self, arr, shape):
        """Reshape 1D arrays into 2D arrays while also handling boundaries.

        :param arr: Array to be reshaped
        :param tuple(int) shape: Shape of the output array


        """
        n0, n1 = shape
        n0 -= 2
        n_bound = n1 // 2
        inner = arr[n_bound:-n_bound].reshape(n0, n1)

        def repeat(values):
            """Repeat an array like [0, 1, 2] -> [0, 0, 1, 1, 2, 2]"""
            n = values.size * 2
            values2 = np.empty((1, n), dtype=values.dtype)
            values2[0, ::2] = values
            values2[0, 1::2] = values
            return values2

        left_boundary = repeat(arr[:n_bound])
        right_boundary = repeat(arr[-n_bound:])

        return np.vstack((left_boundary, inner, right_boundary))

    def plot_contours(self, key, normal=1, interpolate=True, ax=None, **kwargs):
        # No sorting and reshaping is required if interpolation is allowed
        sort_reshape = not interpolate
        x, y, z, inds = self.get_coords(normal, sort_reshape, sort_reshape)
        field = self.get_array(key)

        import matplotlib.pyplot as plt

        if not ax:
            ax = plt

        if normal == 0:
            x0, x1 = y, z
        elif normal == 1:
            x0, x1 = x, z
        elif normal == 2:
            x0, x1 = x, y

        if interpolate:
            ax.tricontourf(x0, x1, field, **kwargs)
        else:
            #  field = field[inds].reshape(x0.shape)
            field = self.reshape(field[inds], shape=x0.shape)
            # print(field.shape, x0.shape)
            ax.contourf(x0, x1, field, **kwargs)


if __name__ in ("__main__", "__vtkconsole__"):
    pv._DisableFirstRenderCameraReset()
    reader = ReaderParaview(
        filename="abl.nek5000",
        arrays=("velocity", "velocity_mag"),
    )

    print(reader.time)
    # next(reader)

    # reader.render()
    #  for (time,) in reader:
    #      # reader.show("velocity_mag")
    #      print(time)
    #      break

    # reader.render()
    slice_data = reader.get_slice(y=0.1)

    velocity_mag = slice_data.get_array("velocity_mag")
    print(velocity_mag, velocity_mag.shape)
    coords = slice_data.get_array("coords")
    print(coords, coords.shape)
    x, y, z = coords.T
    print(x.shape)

    import matplotlib.pyplot as plt

    renderView1 = reader.renderView1
    renderView1.CameraPosition = [
        1.5430590649448526,
        6.463304979594418,
        4.555389168383411,
    ]
    renderView1.CameraFocalPoint = [1.5707999467849734, 0.5, 0.7853999733924866]
    renderView1.CameraViewUp = [
        0.018053943395482774,
        0.5343398685402572,
        -0.8450768959190954,
    ]
    renderView1.CameraParallelScale = 1.8259971497854517

    # plt.ion()
    reader.time = reader.timesteps[-5]
    fig, axes = plt.subplots(2, sharex=True)
    ax1, ax2 = axes.ravel()
    for (ts,) in reader:
        print("time =", ts)
        # renderView1.Update()
        slice_data = reader.get_slice(y=0.1)

        ax1.set_title("interpolated")
        slice_data.plot_contours("velocity_mag", interpolate=True, ax=ax1)
        ax2.set_title(f"actual, time={ts}")
        plt.title(f"time={ts}")
        slice_data.plot_contours("velocity_mag", interpolate=False, ax=ax2)

        # NOTE: Very important to execute reader.show / pv.Show, without which the animation
        # scene is not updated.
        display = reader.show("velocity_mag", slice_data._slice)
        # display = pv.Show(slice_data._slice, renderView1)
        # pv.ColorBy(display, ('POINTS', "velocity_mag"))

        # reader.render()
        # reader.apply()
        plt.pause(0.2)

    plt.pause(1.0)



import matplotlib.pyplot as plt

from snek5000.output.phys_fields import PhysFields
from snek5000.output.readers.pymech_ import ReaderPymech


class ReaderPymechAvg(ReaderPymech):
    """Horizontally and temporally averaged profiles."""

    tag = "pymech_avg"

    def load(self, prefix="", index="all", t_stat=None, **kwargs):
        ds = super().load(prefix, index, **kwargs)

        avg_dims = ("x", "z", "time")
        # Use xarray's sel method to slice along time and mean method to do the
        # averaging
        self.data = ds.sel(time=slice(t_stat, None)).mean(avg_dims)
        return self.data


class PhysFieldsCanonical(PhysFields):
    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        PhysFields._complete_info_solver(info_solver, classes=(ReaderPymechAvg,))

    def plot_mean_vel(self, marker="x"):
        """Plot mean velocity profiles against y-coordinate."""
        fix, ax = plt.subplots()
        data = self.data
        ax.plot("ux", "y", "", marker=marker, label="$U$", data=data)
        ax.plot("uy", "y", "", marker=marker, label="$V$", data=data)
        ax.plot("uz", "y", "", marker=marker, label="$W$", data=data)
        ax.set(ylabel="$y$")
        ax.legend()



import pytest

from snek5000 import load_simul


@pytest.mark.parametrize("session_id", (0, 1))
def test_load_simul_error(sim_data, session_id):
    sim = load_simul(sim_data / f"session_{session_id:02d}")
    assert sim.params.output.session_id == session_id


@pytest.mark.parametrize("session_id", (0, 1))
def test_load_simul_session(sim_data, session_id):
    sim = load_simul(sim_data, session_id=session_id)
    assert sim.params.output.session_id == session_id



from functools import partial

from snakemake import snakemake

make = partial(
    snakemake, snakefile="Snakefile", targets=("earth", "space"), printshellcmds=True
)


def test_resources(capsys):
    # assert make(resources={"alien": 42})
    assert make(overwrite_resources={"space": {"alien": 42}})
    out, err = capsys.readouterr()
    assert "hello alien number 42" in err


if __name__ == "__main__":
    make()
    make(resources={"alien": 42})
    make(overwrite_resources={"space": {"alien": 42}})



import pytest
import xarray as xr
from pymech.neksuite.field import read_header

import snek5000
from snek5000.output import _make_path_session
from snek5000.params import load_params
from snek5000.util.restart import SnekRestartError, get_status, load_for_restart


def test_too_early(sim_data):
    assert get_status(sim_data).code == 425

    with pytest.raises(SnekRestartError, match="425: Too Early"):
        load_for_restart(sim_data)


def test_locked(sim_data):
    locks_dir = sim_data / ".snakemake" / "locks"
    locks_dir.mkdir(parents=True)
    (locks_dir / "lock").touch()

    assert get_status(sim_data).code == 423


def test_ok(sim_data):
    (sim_data / ".snakemake").mkdir()
    session = _make_path_session(sim_data, 1)
    [file.unlink() for file in session.glob("phill*0.f?????")]

    assert get_status(sim_data).code == 200


def test_reset_content(sim_data):
    (sim_data / ".snakemake").mkdir()

    assert get_status(sim_data).code == 205


def test_not_found(sim_data):
    (sim_data / ".snakemake").mkdir()
    (sim_data / "nek5000").unlink()

    assert get_status(sim_data).code == 404


def test_partial_content(sim_data):
    (sim_data / ".snakemake").mkdir()
    _make_path_session(sim_data, 1)
    [restart.unlink() for restart in sim_data.glob("rs6*")]

    assert get_status(sim_data).code == 206


@pytest.mark.parametrize("prefix_dir", ("phill_", "undefined_solver"))
def test_restart_error(tmpdir_factory, prefix_dir):
    tmpdir = tmpdir_factory.mktemp(prefix_dir)
    with pytest.raises(SnekRestartError):
        load_for_restart(tmpdir)


@pytest.mark.slow
def test_restart(sim_executed):
    # In real workflows, to pre load data in preperation for restart, use:
    # >>> sim_executed = snek5000.load()
    fld_file = sim_executed.output.get_field_file()

    header = read_header(fld_file)
    params, Simul = load_for_restart(
        sim_executed.path_run, use_start_from=fld_file.name
    )

    assert params.output.HAS_TO_SAVE
    assert not params.NEW_DIR_RESULTS

    t_end = params.nek.general.end_time = header.time + 10 * abs(
        params.nek.general.dt
    )  # In phill for some reason dt is negative

    sim = Simul(params)
    sim.make.exec("run_fg")

    header = read_header(sim.output.get_field_file())
    assert header.time == t_end

    header = read_header(sim.output.get_field_file(t_approx=t_end))
    assert header.time == t_end

    # check if params_simul.xml is updated
    params_in_filesystem = load_params(sim_executed.path_run)
    assert params.output.session_id == params_in_filesystem.output.session_id


@pytest.mark.slow
def test_restart_new_dir_results(sim_executed):
    params, Simul = load_for_restart(
        sim_executed.path_run, use_start_from=True, new_dir_results=True
    )
    assert params.output.HAS_TO_SAVE
    assert params.NEW_DIR_RESULTS
    assert params.nek.general.start_from == "init_state.restart"
    sim = Simul(params)
    fld_file = sim_executed.output.get_field_file()
    sim.create_symlink_start_from_file(fld_file)


@pytest.mark.slow
def test_restart_new_dir_results_checkpoint(sim_cbox_executed):
    params, Simul = load_for_restart(
        sim_cbox_executed.path_run, use_checkpoint=1, new_dir_results=True
    )
    assert params.output.HAS_TO_SAVE
    assert params.NEW_DIR_RESULTS
    assert params.nek.chkpoint.chkp_fnumber == 1
    assert params.nek.chkpoint.read_chkpt is True
    sim = Simul(params)
    sim.create_symlinks_checkpoint_files(sim_cbox_executed.output.path_run)


def test_phys_fields_uninit(sim):
    """Should error if trying to load / get_var without executing init_reader."""
    with pytest.raises(
        RuntimeError, match="The reader and the method has not initialized yet."
    ):
        sim.output.phys_fields.load()

    with pytest.raises(
        RuntimeError, match="The reader and the method has not initialized yet."
    ):
        sim.output.phys_fields.get_var()


@pytest.mark.parametrize("reader", (True, "pymech_stats"))
def test_load_with_phys_fields(sim, reader):
    sim2 = snek5000.load(sim.path_run, reader=reader)
    with pytest.raises(FileNotFoundError):
        sim2.output.phys_fields.load()

    with pytest.raises(FileNotFoundError):
        sim2.output.phys_fields.get_var("ux")


def test_load_wrong_phys_fields_reader(sim):
    with pytest.raises(ValueError, match="params.output.phys_fields.available_readers"):
        snek5000.load(sim.path_run, reader=1.0)


pymech_issue = (
    "Pymech does not support non-box meshes, and snek5000-phill has "
    "such a geometry: https://github.com/eX-Mech/pymech/issues/31"
)


@pytest.mark.slow
@pytest.mark.xfail(NotImplementedError, reason=pymech_issue)
def test_phys_fields_load_all(sim_executed):
    sim_executed.output.phys_fields.init_reader()
    ds = sim_executed.output.phys_fields.load(index="all")
    assert isinstance(ds, xr.Dataset)


@pytest.mark.slow
@pytest.mark.xfail(NotImplementedError, reason=pymech_issue)
def test_phys_fields_load_stats(sim_executed):
    sim_executed.output.phys_fields.change_reader("pymech_stats")
    ds = sim_executed.output.phys_fields.load(index="*0")
    assert isinstance(ds, xr.Dataset)


@pytest.mark.slow
def test_phys_fields(sim_executed):
    sim_executed.output.phys_fields.init_reader()
    try:
        ds = sim_executed.output.phys_fields.load()
    except NotImplementedError:
        pytest.xfail(reason=pymech_issue)
    else:
        assert isinstance(ds, xr.Dataset)



"""Snakemake interface
======================

"""
import argparse
import sys
from pathlib import Path
from typing import Iterable
from warnings import warn

import yaml
from filelock import FileLock
from snakemake import snakemake
from snakemake.executors import change_working_directory as change_dir

import snek5000


def unlock(path_dir):
    """Unlock a directory locked using Snakemake."""
    with change_dir(path_dir):
        snakemake("Snakefile", unlock=True)


class Make:
    """Snakemake interface for the solvers.

    :param sim: A simulation instance

    """

    def __init__(self, sim=None, path_run=None):

        if (sim is None and path_run is None) or (
            sim is not None and path_run is not None
        ):
            raise ValueError("Either sim of path_run has to be given.")

        if path_run is None:
            self.path_run = sim.output.path_run
            try:
                self.file = next(
                    f for f in sim.output.get_paths() if f.name == "Snakefile"
                )
            except AttributeError:
                raise AttributeError("Unable to get path of Snakefile via Output class")
            except StopIteration:
                raise FileNotFoundError(f"No Snakefile in {self.path_run}")
        else:
            if not path_run.exists():
                raise FileNotFoundError(f"{path_run} does not exist.")
            self.path_run = Path(path_run)
            self.file = self.path_run / "Snakefile"
            if not self.file.exists():
                raise FileNotFoundError(f"No Snakefile in {path_run}")

        self.log_handler = []

    def list(self):
        """List rules.

        Equivalent to::

          snakemake --list-target-rules

        """
        with change_dir(self.path_run):
            return snakemake(self.file, listrules=True, log_handler=self.log_handler)

    def exec(
        self,
        *rules,
        set_resources=None,
        dryrun=False,
        keep_incomplete=True,
        nproc=None,
        **kwargs,
    ):
        """Execute snakemake rules in sequence.

        Parameters
        ----------
        rules: iterable of str, positional-only
            Snakemake rules to be executed. Default rule is `"run"`
        set_resources: dict[str, int]
            Resources to override (see example below).
        dryrun: bool
            Dry run snakemake rules without executing
        keep_incomplete: bool
            Keep incomplete output files of failed jobs
        nproc: Optional[int]
            Number of MPI processes

        Notes
        -----

        For more on available keyword arguments refer to `Snakemake API documentation`_.

        .. _Snakemake API documentation: https://snakemake-api.readthedocs.io/en/stable/api_reference/snakemake.html

        Returns
        -------
        bool
            ``True`` if workflow execution was successful.

        Examples
        --------

        >>> sim.make.exec('mesh', 'SESSION.NAME')
        >>> sim.make.exec('compile')
        >>> sim.make.exec('run', set_resources={'nproc': 4})
        >>> sim.make.exec('run_fg', nproc=2)

        It is also possible to do the same directly from command line
        by changing to the simulation directory and executing::

          snek-make -h
          snek-make compile
          snakemake -j1 compile
          snakemake -j1 --set-resources='run:nproc=4' run

        The flag ``-j`` is short for ``--jobs`` and sets the number of
        threads available for snakemake rules to execute.

        The list of the available target rules can be obtained with:

        >>> sim.make.list()

        or from command line with::

          snek-make -l
          snakemake --list-target-rules

        .. seealso:: Useful Snakemake `command line arguments`_

        .. _command line arguments: https://snakemake.readthedocs.io/en/stable/executing/cli.html#useful-command-line-arguments

        """
        # Default rule
        if not rules:
            rules = ("run",)

        if "resources" in kwargs and not set_resources:
            warn(
                (
                    "Setting `resources` does not have the intended effect. "
                    "Converting `resources` to `set_resources`. "
                    "Use `set_resources` instead in the future."
                ),
                DeprecationWarning,
            )
            set_resources = kwargs.pop("resources")

        if nproc is not None:
            if set_resources is None:
                set_resources = {"nproc": nproc}
            else:
                set_resources["nproc"] = nproc

        if set_resources:
            overwrite_resources = {rule: set_resources for rule in rules}
        else:
            overwrite_resources = None

        # Check if rules were passed as a list / tuple (old API)
        first_rule = rules[0]
        if (
            not isinstance(first_rule, str)
            and isinstance(first_rule, Iterable)
            and len(rules) == 1
        ):
            warn(
                (
                    f"Rules {first_rule} should be passed as positional arguments, "
                    "i.e. a string or comma separated strings; and not as a "
                    f"{type(first_rule)}. Changed in snek5000 0.8.0b0."
                ),
                DeprecationWarning,
            )
            rules = first_rule

        with change_dir(self.path_run):
            return snakemake(
                self.file,
                targets=rules,
                dryrun=dryrun,
                keep_incomplete=keep_incomplete,
                log_handler=self.log_handler,
                overwrite_resources=overwrite_resources,
                **kwargs,
            )


class _Nek5000Make(Make):
    """Snakemake interface to build Nek5000 tools and other dependencies.
    This class would prevent unnecessary rebuild of Nek5000 if there is no
    change in the compiler configuration.

    """

    def __init__(self):
        #: ``NEK_SOURCE_ROOT`` is the working directory
        self.path_run = Path(snek5000.get_nek_source_root())

        #: The ``nek5000.smk`` Snakemake file from :mod:`snek5000.resources` is used here
        self.file = snek5000.get_snek_resource("nek5000.smk")

        #: A file lock ``nek5000_make_config.yml.lock`` used to prevent race conditions
        self.lock = FileLock(self.path_run / "nek5000_make_config.yml.lock")

        #: A YAML file ``nek5000_make_config.yml`` to record compiler configuration
        self.config_cache = self.path_run / "nek5000_make_config.yml"

        #: List of files to build
        self.targets = [
            "bin/genbox",
            "bin/genmap",
            "3rd_party/gslib/lib/libgs.a",
            "3rd_party/blasLapack/libblasLapack.a",
        ]

        # TODO: replace with Snek5000 log handler?
        self.log_handler = []

    def has_to_build(self, compiler_config):
        compiler_config_str = yaml.safe_dump(compiler_config)
        compiler_config_cache = (
            self.config_cache.read_text() if self.config_cache.exists() else ""
        )

        if (
            # One or more of the targets do not exist
            any(not (self.path_run / target).exists() for target in self.targets)
            or compiler_config_str != compiler_config_cache
            # New compiler configuration
        ):
            # Save new configuration first
            self.config_cache.write_text(compiler_config_str)

            return True
        else:
            return False

    def build(self, config):
        """Build Nek5000 tools and third party libraries essential for a
        simulation.

        Parameters
        ----------
        config: dict
            Snakemake configuration

        Returns
        -------
        bool
            ``True`` if workflow execution was successful.

        """

        compiler_config = {
            key: config[key]
            for key in (
                "CC",
                "FC",
                "MPICC",
                "MPIFC",
                "CFLAGS",
                "FFLAGS",
            )
        }

        with self.lock:
            # Only one process can inspect at a time. No timeout
            if self.has_to_build(compiler_config):
                return self.exec(*self.targets, config=config, force_incomplete=True)
            else:
                return True


def snek_make():
    """Used for the command snek-make"""
    parser = argparse.ArgumentParser(
        prog="snek-make",
        description="Execute Snakemake rules",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "rule",
        nargs="?",
        default=None,
        help="Snakemake rule to be executed.",
    )
    parser.add_argument(
        "-l",
        "--list-rules",
        action="store_true",
        help="List rules and exit.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Dry run snakemake rules without executing.",
    )
    parser.add_argument(
        "-np",
        "--nproc",
        type=int,
        default=None,
        help="Number of MPI processes.",
    )
    parser.add_argument(
        "--clean-after-fail",
        action="store_true",
        help="Clean incomplete output files of failed jobs.",
    )

    args = parser.parse_args()

    make = Make(path_run=Path.cwd())

    if args.rule is None:
        print("You need to specify a rule to be executed")

    if args.list_rules or args.rule is None:
        make.list()
        sys.exit(0)

    make.exec(
        args.rule,
        dryrun=args.dry_run,
        keep_incomplete=not args.clean_after_fail,
        nproc=args.nproc,
    )



"""Load and parse stdout from Nek5000.

"""
import re
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

from fluiddyn.util.util import modification_date
from snek5000 import mpi

# Uses awkupy
# %load_ext iawk


class PrintStdOut:
    """Parse standard output log files."""

    _tag = "print_stdout"

    def __init__(self, output=None, path_file=None):
        self.output = output
        self._path_file = path_file
        self.data = None
        self._data_modif_time = None

    @property
    def path_file(self):
        output = self.output
        if output and not self._path_file:
            path_run = Path(output.path_run)
            logfiles = sorted(path_run.glob("*.log"))
            if logfiles:
                try:
                    self._path_file = next(
                        file
                        for file in logfiles
                        if file.name == f"{output.name_solver}.log"
                    )
                except StopIteration:
                    self._path_file = logfiles[-1]
            else:
                self._path_file = path_run / f"{output.name_solver}.log"

        return self._path_file.resolve()

    @path_file.setter
    def path_file(self, path_log_file):
        self._path_file = path_log_file

    @property
    def text(self):
        with open(self.path_file) as fp:
            return fp.read()

    def load(self, pressure_solver="gmres"):
        """Load time data from the log file"""
        # Parse text starting with Step
        # https://regex101.com/r/enFOAg/1

        path_file_time = modification_date(self.path_file)
        if self.data is not None and path_file_time <= self._data_modif_time:
            return self.data

        pattern_step = r"""^Step          # For all lines starting with Step
                \s*            # Followed by some whitespaces
                (?P<it>\d+)    # 0: Capture timestep which are integers
                .*             # Followed by some characters until
                t=\s*          # t=spaces
                (?P<t>\S+)     # 1: Capture t: any non-whitespace char
                ,\s*           # comma and spaces
                DT=\s*         # DT=spaces
                (?P<dt>\S+)    # 2: Capture DT: any non-whitespace char
                ,\s*           # comma and spaces
                C=\s*          # C=spaces
                (?P<CFL>\S+)   # 3: Capture C: any non-whitespace char
            """
        # TODO: add undocumented params.nek.pressure.solver parameter
        # See Line 445 in Nek5000/core/reader_par.f
        pattern_pressure = rf"""^\ +
                (?P<it2>\d+)
                .*
                PRES\ {pressure_solver}    # For all lines containing a string like PRES gmres
                \s+
                (?P<pres_it>\d+)
                \s+
                (?P<pres_div>\S+)
                \s+
                (?P<pres_div0>\S+)
                \s+
                (?P<pres_tol>\S+)
                \s+
                (?P<pres_etime>\S+)
                \s+
                (?P<pres_etime1>\S+)
            """

        expr = re.compile(
            f"({pattern_step})|({pattern_pressure})",
            re.VERBOSE | re.MULTILINE,
        )

        keys_all = tuple(expr.groupindex)
        # Divide into two tuples at "it2"
        idx = keys_all.index("it2")
        keys_step, keys_pressure = keys_all[:idx], keys_all[idx:]

        matches = tuple(expr.finditer(self.text))

        def make_dict(keys):
            return {key: [m[key] for m in matches] for key in keys}

        def make_df(keys):
            return (
                pd.DataFrame.from_dict(make_dict(keys))
                .dropna()
                .astype({key: int if key == "it" else float for key in keys})
            )

        df_step = make_df(keys_step).set_index("it")
        df_pressure = (
            make_df(keys_pressure).rename(columns={"it2": "it"}).set_index("it")
        )

        self.data = df_step.join(df_pressure)
        self._data_modif_time = path_file_time
        return self.data

    def __call__(self, *args):
        """Print to stdout and log file simultaneously."""
        mpi.printby0(*args)
        if mpi.rank == 0 and self.output._has_to_save:
            with self.path_file.open("a") as f:
                f.write(" ".join(str(a) for a in args) + "\n")

    def plot_dt_cfl(self):
        """Plot the evolution of the time step and the CFL number"""

        df = self.load()

        fig, (ax_top, ax_bot) = plt.subplots(nrows=2, sharex=True)

        ax_bot.plot(df.t, df.dt)
        ax_bot.set_title("time step")
        ax_bot.set_xlabel("time")

        ax_top.plot(df.t, df.CFL)
        ax_top.set_title("CFL number")

        params = self.output.sim.params

        if params.nek.general.variable_dt:
            ax_top.axhline(params.nek.general.target_cfl, color="r", label="Target CFL")
            ax_top.legend(loc="lower right")

        fig.tight_layout()

    def plot_nb_iterations(self):
        """Plot the evolution of the number of iterations for the Poisson solver"""

        df = self.load()

        fig, ax = plt.subplots()

        ax.plot(df.t, df.pres_it)
        ax.set_ylabel("number of iterations")
        ax.set_xlabel("time")

        fig.tight_layout()



"""IPython magic extensions
===========================

Usage example in an IPython console:

.. code-block:: ipython

   In [1]: %load_ext snek5000.magic

   In [2]: %snek abl
   INFO     Reading baseline parameters from /home/avmo/src/snek5000/snek5000/src/abl/abl.par
   Created Simul class and default parameters for abl -> Simul, params

   In [3]: sim = Simul(params)

"""
from IPython.core.magic import line_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments

from fluidsim_core.magic import MagicsCore


@magics_class
class SnekMagics(MagicsCore):
    """Magic commands can be loaded in IPython or Jupyter as

    >>> %load_ext fluidsim.magic

    Examples
    --------

    - Magic command %fluidsim or %snek (alias)

    %snek creates the variables `params` and `Simul` for a particular solver.

    Create default parameters for a solver:

    >>> %snek phill

    If a variable `params` already exists, you will be ask if you really want to
    overwrite it. To skip this question:

    >>> %snek phill -f

    List all available solvers and initialized simulation:

    >>> %snek

    - Other fluidsim magic commands

    Quick reference (print this help message):

    >>> %fluidsim_help

    Delete the objects `sim` and `params`:

    >>> %fluidsim_reset

    .. todo::

        - Magic command %fluidsim_load

        %fluidsim_load creates the variables `sim`, `params` and `Simul` from an
        existing simulation.

        Load existing simulation excluding state_phys files:

        >>> %fluidsim_load

        Load existing simulation all options: force overwrite, with state_phys
        files, merging parameters:

        >>> %fluidsim_load -f -s -t -m


    """

    entrypoint_grp = "snek5000.solvers"

    @magic_arguments()
    @argument("solver", nargs="?", default="")
    @argument("-f", "--force-overwrite", action="store_true")
    @line_magic
    def snek(self, line):
        super().fluidsim(line)


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    magics = SnekMagics(ipython)
    ipython.register_magics(magics)



"""
Manage outputs

.. autosummary::
   :toctree:

   base
   history_points
   print_stdout
   phys_fields
   spatial_means

"""
from pathlib import Path


def _make_path_session(path_run: str, session_id: int):
    """Create a path object pointing to session directory.

    Parameters
    ----------
    path_run : str
    session_id : int
    """
    return Path(path_run) / f"session_{session_id:02d}"


def _parse_path_run_session_id(path_session):
    """Parse `path_run` and `session_id` from path_session

    Parameters
    ----------
    path_session : str

    Returns
    -------
    path_dir, session_id: tuple[Path, int | None]
        If the path is to a session_directory the `path_run` (its parent path)
        and the `session_id` will be returned. If not the input parameter will
        be returned as a path object unchanged, along with `None` instead
        of `session_id`.

    """
    path_session = Path(path_session).resolve()

    head, underscore, tail = path_session.name.partition("_")
    if head == "session" and underscore == "_" and tail.isdecimal():
        return path_session.parent, int(tail)
    else:
        return path_session, None



# isort: skip_file
"""
API reference

.. rubric:: Sub-packages

.. autosummary::
   :toctree:

   resources
   output
   solvers
   util

.. rubric:: Modules

.. autosummary::
   :toctree:

   clusters
   const
   info
   log
   magic
   make
   operators
   params
   config

"""

from importlib import resources as _resources
import os
import weakref
from pathlib import Path

from fluiddyn.util import mpi  # noqa: F401

from ._version import __version__  # noqa: F401
from .log import logger  # noqa: F401
from .params import load_params


def get_nek_source_root():
    """Path of Nek5000 source code."""
    if os.getenv("SOURCE_ROOT"):
        logger.warning(
            "The environment variable "
            "SOURCE_ROOT is deprecated in v19, use NEK_SOURCE_ROOT instead."
        )

    try:
        NEK_SOURCE_ROOT = os.environ["NEK_SOURCE_ROOT"]
    except KeyError:
        raise RuntimeError(
            "NEK_SOURCE_ROOT has to point to Nek5000 top level directory."
        )

    nek5000 = str(Path(os.path.expandvars(NEK_SOURCE_ROOT)).expanduser().absolute())

    if not os.path.exists(nek5000):
        logger.error("NEK_SOURCE_ROOT: " + nek5000)
        raise IOError(
            "Cannot find Nek5000 source code. Use environment variable "
            "NEK_SOURCE_ROOT to point to Nek5000 top level directory."
        )
    return nek5000


def get_snek_resource(resource_name):
    """Fetches path of a file from subpackage ``snek5000.resources``.

    Parameters
    ----------
    resource_name: str

        Name of a file packaged in :mod:`snek5000.resources`

    """
    try:
        path = next(_resources.path("snek5000.resources", resource_name).gen)
    except AttributeError:
        path = _resources.path("snek5000.resources", resource_name)
    return path


def load_simul(path_dir=".", session_id=None, reader=True):
    """Loads a simulation

    .. todo::

        Now only the parameters are loaded. Load state too.

    Parameters
    ----------
    path_dir: str or path-like
        Path to a directory containing a simulation. If not provided the
        current directory is used.

    session_id: int
        Indicate which session directory should be used to set
        ``sim.output.path_session``. If not specified it would default to the
        `session_id` and `path_session` values last recorded in the
        `params_simul.xml` file

    reader: bool or str
        By default (`reader=True`) invokes
        :meth:`sim.output.phys_fields.init_reader()
        <snek5000.output.phys_fields.PhysFields.init_reader>`. If a string is
        provided, it is passed onto
        :meth:`sim.output.phys_fields.change_reader(reader)
        <snek5000.output.phys_fields.PhysFields.change_reader>`.


    .. hint::

        An common use case of ``session_id`` parameter is to load field
        files from old sessions using the :meth:`sim.output.get_field_file
        <snek5000.output.base.Output.get_field_file>` method

    """
    from snek5000.util.files import _path_try_from_fluidsim_path

    path_dir = _path_try_from_fluidsim_path(path_dir)

    # Load simulation class
    from snek5000.solvers import import_cls_simul, get_solver_short_name
    from snek5000.output import _parse_path_run_session_id

    # In case the user specifies the path to a session sub-directory
    if session_id is None:
        path_dir, session_id = _parse_path_run_session_id(path_dir)

    short_name = get_solver_short_name(path_dir)
    Simul = import_cls_simul(short_name)

    # Load parameters
    from snek5000.params import load_params

    params = load_params(path_dir)

    # Modify parameters prior to loading
    params.NEW_DIR_RESULTS = False
    params.output.HAS_TO_SAVE = False
    params.path_run = path_dir

    if isinstance(session_id, int):
        from snek5000.output import _make_path_session

        params.output.session_id = session_id
        params.output.path_session = _make_path_session(path_dir, session_id)

    sim = Simul(params)

    if reader:
        if reader is True:
            sim.output.phys_fields.init_reader()
        elif isinstance(reader, str):
            sim.output.phys_fields.change_reader(reader)
        else:
            raise ValueError(
                f"Reader should be either True or False or one of {params.output.phys_fields.available_readers = }"
            )

    return sim


# for consistency with fluidsim
load = weakref.proxy(load_simul)  #: Alias for :func:`load_simul`


from .util.smake import append_debug_flags, ensure_env  # noqa: F401, E402
from .util import restart  # noqa: F401, E402

get_status = restart.get_status  #: Alias for :func:`snek5000.util.restart.get_status`
load_for_restart = (
    restart.load_for_restart
)  #: Alias for :func:`snek5000.util.restart.load_for_restart`


__all__ = [
    "get_nek_source_root",
    "get_snek_resource",
    "get_status",
    "load",
    "load_for_restart",
    "load_params",
    "load_simul",
    "logger",
]



import pytest

try:
    from IPython.testing.globalipapp import get_ipython

    ip = get_ipython()
    ip.run_line_magic("load_ext", "snek5000.magic")
except ImportError:
    ip = False


@pytest.mark.skipif(not ip, reason="Magics cannot be tested if IPython is unavailable")
def test_snek5000_magic():
    ip.run_line_magic("snek", "kth")



"""Cluster utilities"""
import os

from .log import logger


def _int_getenv(key, default=0):
    return int(os.getenv(key, default))


def _slurm():
    """Check different environment variables available within a SLURM job.

    .. seealso::

        https://slurm.schedmd.com/srun.html#SECTION_OUTPUT-ENVIRONMENT-VARIABLES

    """
    return _int_getenv("SLURM_NPROCS") or (
        _int_getenv("SLURM_NNODES") * _int_getenv("SLURM_CPUS_ON_NODE")
    )


def _oar():
    """Equivalent to  the shell command::

    NPROC=$(cat ${OAR_NODEFILE} | wc -l).

    """
    nodefile = os.getenv("OAR_NODEFILE")
    if nodefile and os.path.exists(nodefile):
        with open(nodefile) as file:
            return len(file.readlines())
    else:
        return 0


def nproc_available():
    """Detect maximum number of processors available

    Returns
    -------
    int

    """
    # Inside job schedulers, number of processors assigned will vary
    # TODO: extend this function for other schedulers PBS etc.
    cluster_nproc = _slurm() or _oar()

    if cluster_nproc:
        var = "Cluster environment variables"
        nproc = cluster_nproc
    else:
        var = "os.cpu_count()"
        nproc = os.cpu_count()

    logger.debug(f"Using {var} to detect maximum number of processors available")
    return nproc



import pytest

from snek5000 import load_simul
from snek5000.solvers import available_solvers, import_cls_simul


def test_entrypoints():
    solvers = [entrypoint.name for entrypoint in available_solvers()]
    assert "nek" in solvers
    assert "kth" in solvers
    assert "phill" in solvers


def test_init_base():
    from snek5000.solvers.base import Simul

    params = Simul.create_default_params()
    Simul(params)


def test_init_kth():
    from snek5000.solvers.kth import Simul

    params = Simul.create_default_params()
    Simul(params)


def test_import_solver():
    from phill.solver import Simul

    assert Simul is import_cls_simul("phill")
    params = Simul.create_default_params()
    sim1 = Simul(params)
    sim2 = load_simul(sim1.path_run)
    assert isinstance(sim1, Simul) and isinstance(sim2, Simul)


def test_output(sim_data):
    sim = load_simul(sim_data)

    assert (
        sim.output.print_stdout.path_file.exists()
    ), "Cannot find .log file for print_stdout"
    try:
        sim.output.phys_fields
    except AttributeError as err:
        raise AssertionError("Cannot initialize phys_fields") from err


def test_make_exec_deprecation(sim2d):
    with pytest.deprecated_call():
        sim2d.make.exec(["phill.re2", "phill.ma2"])

    sim2d.make.exec("SESSION.NAME")


def test_init_output():
    from phill.solver import Simul

    params = Simul.create_default_params()

    # init history_points
    coords = params.output.history_points.coords = [
        (0.5, 0.2, 0.5),
        (0.5, 0.8, 0.5),
    ]
    params.oper.max.hist = len(coords) + 1

    sim = Simul(params)

    # tests history_points
    assert sim.output.history_points.path_file.exists()
    coords, df = sim.output.history_points.load()

    assert len(df) == 0
    assert tuple(coords.iloc[-1]) == params.output.history_points.coords[-1]

    with open(sim.output.history_points.path_file, "a") as file:
        file.write(
            "\n".join(
                ["0.0 2.0 2.0 2.0 2.0"] * 2
                + ["0.5 2.0 2.0 2.0 2.0"] * 2
                + ["1.0 2.0 2.0 2.0 2.0\n2.0 2.0\n"]
            )
        )

    sim.output.history_points.plot()



from snek5000.info import InfoSolverMake
from snek5000.solvers.base import SimulNek

# To use KTH Framework import SimulKTH instead
# from snek5000.solvers.kth import SimulKTH


class InfoSolverCanonical(InfoSolverMake):
    """Contain the information on a :class:`snek5000_canonical.solver.Simul`
    instance.

    """

    def _init_root(self):
        from . import short_name

        super()._init_root()
        self.module_name = "snek5000_canonical.solver"
        self.class_name = "Simul"
        self.short_name = short_name

        self.classes.Output.module_name = "snek5000_canonical.output"
        self.classes.Output.class_name = "OutputCanonical"

        # To solve for the temperature, one needs to add a [TEMPERATURE]
        # section in the .par file. It can be done like this:
        self.par_sections_disabled.remove("temperature")


class SimulCanonical(SimulNek):
    """A solver which compiles and runs using a Snakefile."""

    InfoSolver = InfoSolverCanonical

    @classmethod
    def _complete_params_with_default(cls, params):
        """Add missing default parameters."""
        params = super()._complete_params_with_default(params)
        # Extend with new default parameters here, for example:

        # params.nek.velocity._set_attrib("advection", True)
        return params

    @classmethod
    def create_default_params(cls):
        """Set default values of parameters as given in reference
        implementation.

        """
        params = super().create_default_params()
        # Re-define default values for parameters here, if necessary
        # following ``canonical.par``, ``canonical.box`` and ``SIZE`` files
        return params


Simul = SimulCanonical
